"""Recursive PLS (RPLS) regressor for nirs4all.

A sklearn-compatible implementation of Recursive Partial Least Squares.
RPLS enables online model updates for drifting processes through incremental
updates using a forgetting factor.

Supports both NumPy (CPU) and JAX (GPU/TPU) backends.

References
----------
- Qin, S. J. (1998). Recursive PLS algorithms for adaptive data modeling.
  Computers & Chemical Engineering, 22(4-5), 503-514.
- Helland, K., Berntsen, H. E., Borgen, O. S., & Martens, H. (1992).
  Recursive algorithm for partial least squares regression.
  Chemometrics and Intelligent Laboratory Systems, 14(1-3), 129-137.
- Dayal, B. S., & MacGregor, J. F. (1997). Recursive exponentially
  weighted PLS and its applications to adaptive control and prediction.
  Journal of Process Control, 7(3), 169-179.
"""

from __future__ import annotations

from functools import partial
from typing import Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted


def _check_jax_available():
    """Check if JAX is available for GPU acceleration."""
    try:
        import jax
        return True
    except ImportError:
        return False

# =============================================================================
# NumPy Backend Implementation
# =============================================================================

def _initial_pls_fit_numpy(
    X: NDArray[np.floating],
    Y: NDArray[np.floating],
    n_components: int,
) -> tuple[
    NDArray[np.floating],  # W (X weights)
    NDArray[np.floating],  # P (X loadings)
    NDArray[np.floating],  # Q (Y loadings)
    NDArray[np.floating],  # R (W * inv(P'W) for direct regression)
    NDArray[np.floating],  # B (regression coefficients)
    NDArray[np.floating],  # Cov_X (running covariance of X)
    NDArray[np.floating],  # Cov_XY (running cross-covariance of X and Y)
]:
    """Initial batch PLS fit using SIMPLS-like algorithm.

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features)
        Centered X matrix.
    Y : ndarray of shape (n_samples, n_targets)
        Centered Y matrix.
    n_components : int
        Number of PLS components.

    Returns
    -------
    W, P, Q, R, B, Cov_X, Cov_XY : tuple of ndarrays
    """
    n_samples, n_features = X.shape
    n_targets = Y.shape[1]

    # Initialize arrays
    W = np.zeros((n_features, n_components), dtype=np.float64)
    P = np.zeros((n_features, n_components), dtype=np.float64)
    Q = np.zeros((n_targets, n_components), dtype=np.float64)
    V = np.zeros((n_features, n_components), dtype=np.float64)

    # Initial covariance matrices
    Cov_X = X.T @ X  # (n_features, n_features)
    Cov_XY = X.T @ Y  # (n_features, n_targets)
    S = Cov_XY.copy()

    for a in range(n_components):
        # Get weight as dominant left singular vector of S
        if n_targets == 1:
            w = S[:, 0].copy()
        else:
            u_svd, s_svd, vh = np.linalg.svd(S, full_matrices=False)
            w = u_svd[:, 0]

        # Normalize
        w_norm = np.linalg.norm(w)
        if w_norm < 1e-14:
            break
        w = w / w_norm

        # Score and loadings
        t = X @ w
        tt = t.T @ t
        if tt < 1e-14:
            break

        p = (X.T @ t) / tt
        q = (Y.T @ t) / tt

        # Store
        W[:, a] = w
        P[:, a] = p.ravel()
        Q[:, a] = q.ravel()

        # Orthogonalization for deflation
        v = p.ravel().copy()
        if a > 0:
            v = v - V[:, :a] @ (V[:, :a].T @ p.ravel())
        v_norm = np.linalg.norm(v)
        if v_norm > 1e-14:
            v = v / v_norm
        V[:, a] = v

        # Deflate S
        S = S - v[:, np.newaxis] @ (v[np.newaxis, :] @ S)

    # Compute R = W @ inv(P'W)
    PtW = P.T @ W
    try:
        R = W @ np.linalg.inv(PtW)
    except np.linalg.LinAlgError:
        R = W @ np.linalg.pinv(PtW)

    # Compute regression coefficients
    B = R @ Q.T

    return W, P, Q, R, B, Cov_X, Cov_XY

def _recursive_update_numpy(
    X_new: NDArray[np.floating],
    Y_new: NDArray[np.floating],
    n_components: int,
    W: NDArray[np.floating],
    P: NDArray[np.floating],
    Q: NDArray[np.floating],
    Cov_X: NDArray[np.floating],
    Cov_XY: NDArray[np.floating],
    forgetting_factor: float,
    n_samples_seen: int,
) -> tuple[
    NDArray[np.floating],  # W
    NDArray[np.floating],  # P
    NDArray[np.floating],  # Q
    NDArray[np.floating],  # R
    NDArray[np.floating],  # B
    NDArray[np.floating],  # Cov_X
    NDArray[np.floating],  # Cov_XY
    int,  # n_samples_seen
]:
    """Recursive update of PLS model.

    Uses exponentially weighted recursive least squares to update
    the covariance matrices and recompute PLS loadings.

    Parameters
    ----------
    X_new : ndarray of shape (n_new_samples, n_features)
        New centered X samples.
    Y_new : ndarray of shape (n_new_samples, n_targets)
        New centered Y samples.
    n_components : int
        Number of PLS components.
    W, P, Q : ndarrays
        Current weights and loadings.
    Cov_X, Cov_XY : ndarrays
        Current covariance matrices.
    forgetting_factor : float
        Forgetting factor in (0, 1]. Lower values forget faster.
    n_samples_seen : int
        Number of samples seen so far.

    Returns
    -------
    Updated W, P, Q, R, B, Cov_X, Cov_XY, n_samples_seen
    """
    n_new = X_new.shape[0]
    n_features = X_new.shape[1]
    n_targets = Y_new.shape[1]

    ff = forgetting_factor

    # Update covariance matrices with exponential weighting
    for i in range(n_new):
        x_i = X_new[i:i+1].T  # (n_features, 1)
        y_i = Y_new[i:i+1].T  # (n_targets, 1)

        # Exponentially weighted update
        Cov_X = ff * Cov_X + x_i @ x_i.T
        Cov_XY = ff * Cov_XY + x_i @ y_i.T
        n_samples_seen += 1

    # Recompute PLS components from updated covariances
    # This is the key step: use updated covariances to get new loadings

    V = np.zeros((n_features, n_components), dtype=np.float64)
    S = Cov_XY.copy()

    for a in range(n_components):
        # Get weight from covariance
        if n_targets == 1:
            w = S[:, 0].copy()
        else:
            u_svd, s_svd, vh = np.linalg.svd(S, full_matrices=False)
            w = u_svd[:, 0]

        w_norm = np.linalg.norm(w)
        if w_norm < 1e-14:
            break
        w = w / w_norm

        # Loading from covariance: p = Cov_X @ w / (w' @ Cov_X @ w)
        Cov_X_w = Cov_X @ w
        denominator = w.T @ Cov_X_w
        if np.abs(denominator) < 1e-14:
            break

        p = Cov_X_w / denominator
        q = Cov_XY.T @ w / denominator

        W[:, a] = w
        P[:, a] = p
        Q[:, a] = q

        # Orthogonalization
        v = p.copy()
        if a > 0:
            v = v - V[:, :a] @ (V[:, :a].T @ p)
        v_norm = np.linalg.norm(v)
        if v_norm > 1e-14:
            v = v / v_norm
        V[:, a] = v

        # Deflate S
        S = S - v[:, np.newaxis] @ (v[np.newaxis, :] @ S)

    # Compute R and B
    PtW = P.T @ W
    try:
        R = W @ np.linalg.inv(PtW)
    except np.linalg.LinAlgError:
        R = W @ np.linalg.pinv(PtW)

    B = R @ Q.T

    return W, P, Q, R, B, Cov_X, Cov_XY, n_samples_seen

# =============================================================================
# JAX Backend Implementation
# =============================================================================

def _get_jax_recursive_pls_functions():
    """Lazy import and create JAX Recursive PLS functions."""
    import jax
    import jax.numpy as jnp
    from jax import lax

    # Enable float64 for numerical precision
    jax.config.update("jax_enable_x64", True)

    @partial(jax.jit, static_argnums=(2,))
    def initial_pls_fit_jax(
        X: jax.Array,
        Y: jax.Array,
        n_components: int,
    ) -> tuple:
        """JIT-compiled initial PLS fit using JAX."""
        n_samples, n_features = X.shape
        n_targets = Y.shape[1]

        # Initial covariance matrices
        Cov_X = X.T @ X
        Cov_XY = X.T @ Y
        S = Cov_XY.copy()

        def component_step(a, carry):
            S, W, P, Q, V = carry

            # Get weight from SVD
            u_svd, s_svd, vh = jnp.linalg.svd(S, full_matrices=False)
            w = u_svd[:, 0]

            # Normalize
            w_norm = jnp.linalg.norm(w)
            w = jnp.where(w_norm > 1e-14, w / w_norm, w)

            # Score and loadings
            t = X @ w
            tt = t.T @ t
            tt_safe = jnp.where(tt > 1e-14, tt, 1.0)

            p = (X.T @ t) / tt_safe
            q = (Y.T @ t) / tt_safe

            # Store
            W = W.at[:, a].set(w)
            P = P.at[:, a].set(p)
            Q = Q.at[:, a].set(q)

            # Orthogonalization
            v = p.copy()
            prev_V = V * jnp.where(jnp.arange(n_components) < a, 1.0, 0.0)
            v = v - prev_V @ (prev_V.T @ p)
            v_norm = jnp.linalg.norm(v)
            v = jnp.where(v_norm > 1e-14, v / v_norm, v)
            V = V.at[:, a].set(v)

            # Deflate
            S = S - jnp.outer(v, v @ S)

            return S, W, P, Q, V

        # Initialize
        W = jnp.zeros((n_features, n_components), dtype=jnp.float64)
        P = jnp.zeros((n_features, n_components), dtype=jnp.float64)
        Q = jnp.zeros((n_targets, n_components), dtype=jnp.float64)
        V = jnp.zeros((n_features, n_components), dtype=jnp.float64)

        init_carry = (S, W, P, Q, V)
        S, W, P, Q, V = lax.fori_loop(0, n_components, component_step, init_carry)

        # Compute R and B
        PtW = P.T @ W
        R = W @ jnp.linalg.pinv(PtW)
        B = R @ Q.T

        return W, P, Q, R, B, Cov_X, Cov_XY

    @partial(jax.jit, static_argnums=(2,))
    def recursive_update_jax(
        X_new: jax.Array,
        Y_new: jax.Array,
        n_components: int,
        W: jax.Array,
        P: jax.Array,
        Q: jax.Array,
        Cov_X: jax.Array,
        Cov_XY: jax.Array,
        forgetting_factor: float,
    ) -> tuple:
        """JIT-compiled recursive PLS update using JAX."""
        n_new = X_new.shape[0]
        n_features = X_new.shape[1]
        n_targets = Y_new.shape[1]

        ff = forgetting_factor

        # Update covariances in a loop using lax.dynamic_slice for JAX compatibility
        def update_cov_step(i, carry):
            Cov_X, Cov_XY = carry
            # Use lax.dynamic_slice instead of Python slicing
            x_i = lax.dynamic_slice(X_new, (i, 0), (1, n_features)).T
            y_i = lax.dynamic_slice(Y_new, (i, 0), (1, n_targets)).T
            Cov_X = ff * Cov_X + x_i @ x_i.T
            Cov_XY = ff * Cov_XY + x_i @ y_i.T
            return Cov_X, Cov_XY

        Cov_X, Cov_XY = lax.fori_loop(0, n_new, update_cov_step, (Cov_X, Cov_XY))

        # Recompute PLS from updated covariances
        S = Cov_XY.copy()

        def component_step(a, carry):
            S, W, P, Q, V = carry

            # Get weight
            u_svd, s_svd, vh = jnp.linalg.svd(S, full_matrices=False)
            w = u_svd[:, 0]

            w_norm = jnp.linalg.norm(w)
            w = jnp.where(w_norm > 1e-14, w / w_norm, w)

            # Loading from covariance
            Cov_X_w = Cov_X @ w
            denom = w.T @ Cov_X_w
            denom_safe = jnp.where(jnp.abs(denom) > 1e-14, denom, 1.0)

            p = Cov_X_w / denom_safe
            q = Cov_XY.T @ w / denom_safe

            W = W.at[:, a].set(w)
            P = P.at[:, a].set(p)
            Q = Q.at[:, a].set(q)

            # Orthogonalization
            v = p.copy()
            prev_V = V * jnp.where(jnp.arange(n_components) < a, 1.0, 0.0)
            v = v - prev_V @ (prev_V.T @ p)
            v_norm = jnp.linalg.norm(v)
            v = jnp.where(v_norm > 1e-14, v / v_norm, v)
            V = V.at[:, a].set(v)

            # Deflate
            S = S - jnp.outer(v, v @ S)

            return S, W, P, Q, V

        V = jnp.zeros((n_features, n_components), dtype=jnp.float64)
        init_carry = (S, W, P, Q, V)
        S, W, P, Q, V = lax.fori_loop(0, n_components, component_step, init_carry)

        # Compute R and B
        PtW = P.T @ W
        R = W @ jnp.linalg.pinv(PtW)
        B = R @ Q.T

        return W, P, Q, R, B, Cov_X, Cov_XY

    @jax.jit
    def predict_jax(
        X: jax.Array,
        X_mean: jax.Array,
        X_std: jax.Array,
        Y_mean: jax.Array,
        Y_std: jax.Array,
        B: jax.Array,
    ) -> jax.Array:
        """JIT-compiled prediction using JAX."""
        X_centered = (X - X_mean) / X_std
        Y_pred_std = X_centered @ B
        Y_pred = Y_pred_std * Y_std + Y_mean
        return Y_pred

    return initial_pls_fit_jax, recursive_update_jax, predict_jax

# Cache for JAX functions
_JAX_RECURSIVE_PLS_FUNCS = None

def _get_cached_jax_recursive_pls():
    """Get cached JAX Recursive PLS functions."""
    global _JAX_RECURSIVE_PLS_FUNCS
    if _JAX_RECURSIVE_PLS_FUNCS is None:
        _JAX_RECURSIVE_PLS_FUNCS = _get_jax_recursive_pls_functions()
    return _JAX_RECURSIVE_PLS_FUNCS

# =============================================================================
# RecursivePLS Estimator Class
# =============================================================================

class RecursivePLS(BaseEstimator, RegressorMixin):
    """Recursive Partial Least Squares (Recursive PLS) regressor.

    Recursive PLS enables online model updates for drifting processes.
    It uses a forgetting factor to exponentially weight old samples,
    allowing the model to adapt to non-stationary data streams.

    The algorithm maintains running covariance matrices that are updated
    incrementally with each new batch of samples. The PLS loadings are
    then recomputed from these updated covariances.

    Parameters
    ----------
    n_components : int, default=10
        Number of PLS components to extract.
    forgetting_factor : float, default=0.99
        Forgetting factor in (0, 1]. Controls the rate of adaptation:
        - 1.0: No forgetting, standard batch PLS
        - <1.0: Exponential forgetting of old samples
        - Typical values: 0.95-0.999 depending on drift speed
    scale : bool, default=True
        Whether to scale X and Y to unit variance.
    center : bool, default=True
        Whether to center X and Y (subtract mean).
    backend : str, default='numpy'
        Computational backend to use:
        - 'numpy': NumPy backend (CPU only).
        - 'jax': JAX backend (supports GPU/TPU acceleration).

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during fit.
    n_components_ : int
        Actual number of components used.
    n_samples_seen_ : int
        Total number of samples seen (including partial_fit calls).
    x_mean_ : ndarray of shape (n_features,)
        Mean of X (updated with exponential moving average).
    x_std_ : ndarray of shape (n_features,)
        Standard deviation of X.
    y_mean_ : ndarray of shape (n_targets,)
        Mean of Y (updated with exponential moving average).
    y_std_ : ndarray of shape (n_targets,)
        Standard deviation of Y.
    x_weights_ : ndarray of shape (n_features, n_components_)
        X weights (W).
    x_loadings_ : ndarray of shape (n_features, n_components_)
        X loadings (P).
    y_loadings_ : ndarray of shape (n_targets, n_components_)
        Y loadings (Q).
    coef_ : ndarray of shape (n_features, n_targets)
        Regression coefficients.

    Examples
    --------
    >>> from nirs4all.operators.models.sklearn.recursive_pls import RecursivePLS
    >>> import numpy as np
    >>> # Initial batch fit
    >>> np.random.seed(42)
    >>> X_init = np.random.randn(100, 50)
    >>> y_init = X_init[:, :5].sum(axis=1) + 0.1 * np.random.randn(100)
    >>> model = RecursivePLS(n_components=10, forgetting_factor=0.99)
    >>> model.fit(X_init, y_init)
    RecursivePLS(n_components=10)
    >>> # Online update with new samples
    >>> X_new = np.random.randn(10, 50)
    >>> y_new = X_new[:, :5].sum(axis=1) + 0.1 * np.random.randn(10)
    >>> model.partial_fit(X_new, y_new)
    >>> # Predict
    >>> predictions = model.predict(X_new)
    >>> print(f"Samples seen: {model.n_samples_seen_}")

    Notes
    -----
    Recursive PLS is particularly useful when:
    - Data arrives in streams and batch retraining is too expensive
    - Process conditions drift over time (sensor aging, raw material changes)
    - You need to adapt a calibration model to local conditions

    The forgetting factor controls the adaptation speed:
    - Higher values (0.99-0.999): Slow adaptation, stable model
    - Lower values (0.9-0.95): Fast adaptation, may be unstable

    See Also
    --------
    SIMPLS : Batch SIMPLS algorithm.
    sklearn.cross_decomposition.PLSRegression : sklearn's batch PLS.

    References
    ----------
    - Qin, S. J. (1998). Recursive PLS algorithms for adaptive data
      modeling. Computers & Chemical Engineering, 22(4-5), 503-514.
    - Dayal, B. S., & MacGregor, J. F. (1997). Recursive exponentially
      weighted PLS and its applications to adaptive control and prediction.
      Journal of Process Control, 7(3), 169-179.
    """

    _webapp_meta = {
        "category": "pls",
        "tier": "advanced",
        "tags": ["pls", "recursive", "online", "adaptive", "regression"],
    }

    # Explicitly declare estimator type for sklearn compatibility (e.g., StackingRegressor)
    _estimator_type = "regressor"

    def __init__(
        self,
        n_components: int = 10,
        forgetting_factor: float = 0.99,
        scale: bool = True,
        center: bool = True,
        backend: str = 'numpy',
    ):
        """Initialize Recursive PLS regressor."""
        self.n_components = n_components
        self.forgetting_factor = forgetting_factor
        self.scale = scale
        self.center = center
        self.backend = backend

    def fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
    ) -> RecursivePLS:
        """Fit the Recursive PLS model with initial batch.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : RecursivePLS
            Fitted estimator.

        Raises
        ------
        ValueError
            If backend is not 'numpy' or 'jax'.
            If forgetting_factor is not in (0, 1].
        ImportError
            If backend is 'jax' and JAX is not installed.
        """
        # Validate backend
        if self.backend not in ('numpy', 'jax'):
            raise ValueError(
                f"backend must be 'numpy' or 'jax', got '{self.backend}'"
            )

        if self.backend == 'jax' and not _check_jax_available():
            raise ImportError(
                "JAX is required for RecursivePLS with backend='jax'. "
                "Install it with: pip install jax\n"
                "For GPU support: pip install jax[cuda12]"
            )

        # Validate forgetting factor
        if not (0 < self.forgetting_factor <= 1):
            raise ValueError(
                f"forgetting_factor must be in (0, 1], got {self.forgetting_factor}"
            )

        X = np.asarray(X, dtype=np.float64)
        y = np.asarray(y, dtype=np.float64)

        # Handle 1D y
        self._y_1d = y.ndim == 1
        if self._y_1d:
            y = y.reshape(-1, 1)

        n_samples, n_features = X.shape
        n_targets = y.shape[1]

        self.n_features_in_ = n_features
        self.n_samples_seen_ = n_samples

        # Limit components
        max_components = min(n_samples - 1, n_features)
        self.n_components_ = min(self.n_components, max_components)

        # Compute and store scaling parameters
        if self.center:
            self.x_mean_ = X.mean(axis=0)
            self.y_mean_ = y.mean(axis=0)
        else:
            self.x_mean_ = np.zeros(n_features, dtype=np.float64)
            self.y_mean_ = np.zeros(n_targets, dtype=np.float64)

        if self.scale:
            self.x_std_ = X.std(axis=0, ddof=1)
            self.y_std_ = y.std(axis=0, ddof=1)
            self.x_std_ = np.where(self.x_std_ < 1e-10, 1.0, self.x_std_)
            self.y_std_ = np.where(self.y_std_ < 1e-10, 1.0, self.y_std_)
        else:
            self.x_std_ = np.ones(n_features, dtype=np.float64)
            self.y_std_ = np.ones(n_targets, dtype=np.float64)

        X_centered = (X - self.x_mean_) / self.x_std_
        Y_centered = (y - self.y_mean_) / self.y_std_

        # Initial fit
        if self.backend == 'jax':
            import jax.numpy as jnp

            fit_jax, _, _ = _get_cached_jax_recursive_pls()

            X_jax = jnp.asarray(X_centered)
            Y_jax = jnp.asarray(Y_centered)

            W, P, Q, R, B, Cov_X, Cov_XY = fit_jax(
                X_jax, Y_jax, self.n_components_
            )

            self.x_weights_ = np.asarray(W)
            self.x_loadings_ = np.asarray(P)
            self.y_loadings_ = np.asarray(Q)
            self._R = np.asarray(R)
            self._B = np.asarray(B)
            self._Cov_X = np.asarray(Cov_X)
            self._Cov_XY = np.asarray(Cov_XY)
        else:
            W, P, Q, R, B, Cov_X, Cov_XY = _initial_pls_fit_numpy(
                X_centered, Y_centered, self.n_components_
            )

            self.x_weights_ = W
            self.x_loadings_ = P
            self.y_loadings_ = Q
            self._R = R
            self._B = B
            self._Cov_X = Cov_X
            self._Cov_XY = Cov_XY

        # Store final coefficients in original space
        self.coef_ = self._B * self.y_std_[np.newaxis, :] / self.x_std_[:, np.newaxis]

        return self

    def partial_fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
    ) -> RecursivePLS:
        """Update the Recursive PLS model with new samples.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            New training data.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            New target values.

        Returns
        -------
        self : RecursivePLS
            Updated estimator.

        Raises
        ------
        NotFittedError
            If the model has not been fitted yet.
        """
        check_is_fitted(self, ['x_mean_', 'x_std_', '_B', '_Cov_X', '_Cov_XY'])

        X = np.asarray(X, dtype=np.float64)
        y = np.asarray(y, dtype=np.float64)

        if self._y_1d:
            y = y.reshape(-1, 1)

        n_new = X.shape[0]

        # Update running mean with exponential moving average
        ff = self.forgetting_factor
        for _ in range(n_new):
            # EMA update for mean (optional: can also keep fixed from initial fit)
            # self.x_mean_ = ff * self.x_mean_ + (1 - ff) * X[i]
            # self.y_mean_ = ff * self.y_mean_ + (1 - ff) * y[i]
            pass  # Keep mean fixed for stability

        # Center new data using stored parameters
        X_centered = (X - self.x_mean_) / self.x_std_
        Y_centered = (y - self.y_mean_) / self.y_std_

        # Recursive update
        if self.backend == 'jax':
            import jax.numpy as jnp

            _, update_jax, _ = _get_cached_jax_recursive_pls()

            X_jax = jnp.asarray(X_centered)
            Y_jax = jnp.asarray(Y_centered)
            W_jax = jnp.asarray(self.x_weights_)
            P_jax = jnp.asarray(self.x_loadings_)
            Q_jax = jnp.asarray(self.y_loadings_)
            Cov_X_jax = jnp.asarray(self._Cov_X)
            Cov_XY_jax = jnp.asarray(self._Cov_XY)

            W, P, Q, R, B, Cov_X, Cov_XY = update_jax(
                X_jax, Y_jax, self.n_components_,
                W_jax, P_jax, Q_jax, Cov_X_jax, Cov_XY_jax,
                self.forgetting_factor
            )

            self.x_weights_ = np.asarray(W)
            self.x_loadings_ = np.asarray(P)
            self.y_loadings_ = np.asarray(Q)
            self._R = np.asarray(R)
            self._B = np.asarray(B)
            self._Cov_X = np.asarray(Cov_X)
            self._Cov_XY = np.asarray(Cov_XY)
        else:
            W, P, Q, R, B, Cov_X, Cov_XY, n_seen = _recursive_update_numpy(
                X_centered, Y_centered, self.n_components_,
                self.x_weights_, self.x_loadings_, self.y_loadings_,
                self._Cov_X, self._Cov_XY,
                self.forgetting_factor, self.n_samples_seen_
            )

            self.x_weights_ = W
            self.x_loadings_ = P
            self.y_loadings_ = Q
            self._R = R
            self._B = B
            self._Cov_X = Cov_X
            self._Cov_XY = Cov_XY

        self.n_samples_seen_ += n_new

        # Update coefficients
        self.coef_ = self._B * self.y_std_[np.newaxis, :] / self.x_std_[:, np.newaxis]

        return self

    def predict(
        self,
        X: ArrayLike,
    ) -> NDArray[np.floating]:
        """Predict using the Recursive PLS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples to predict.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,) or (n_samples, n_targets)
            Predicted values.
        """
        check_is_fitted(self, ['x_mean_', 'x_std_', 'y_mean_', 'y_std_', '_B'])

        X = np.asarray(X, dtype=np.float64)

        if self.backend == 'jax':
            import jax.numpy as jnp

            _, _, predict_jax = _get_cached_jax_recursive_pls()

            X_jax = jnp.asarray(X)
            x_mean_jax = jnp.asarray(self.x_mean_)
            x_std_jax = jnp.asarray(self.x_std_)
            y_mean_jax = jnp.asarray(self.y_mean_)
            y_std_jax = jnp.asarray(self.y_std_)
            B_jax = jnp.asarray(self._B)

            y_pred = predict_jax(
                X_jax, x_mean_jax, x_std_jax, y_mean_jax, y_std_jax, B_jax
            )
            y_pred = np.asarray(y_pred)
        else:
            X_centered = (X - self.x_mean_) / self.x_std_
            y_pred_std = X_centered @ self._B
            y_pred = y_pred_std * self.y_std_ + self.y_mean_

        if self._y_1d:
            y_pred = y_pred.ravel()

        return np.asarray(y_pred)

    def transform(
        self,
        X: ArrayLike,
    ) -> NDArray[np.floating]:
        """Transform X to score space.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples to transform.

        Returns
        -------
        T : ndarray of shape (n_samples, n_components_)
            X scores.
        """
        check_is_fitted(self, ['x_mean_', 'x_std_', '_R'])

        X = np.asarray(X, dtype=np.float64)
        X_centered = (X - self.x_mean_) / self.x_std_

        T = X_centered @ self._R

        return np.asarray(T)

    def get_params(self, deep: bool = True) -> dict:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        return {
            'n_components': self.n_components,
            'forgetting_factor': self.forgetting_factor,
            'scale': self.scale,
            'center': self.center,
            'backend': self.backend,
        }

    def set_params(self, **params) -> RecursivePLS:
        """Set the parameters of this estimator.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : RecursivePLS
            Estimator instance.
        """
        for key, value in params.items():
            setattr(self, key, value)
        return self

    def __repr__(self) -> str:
        """Return string representation."""
        return (
            f"RecursivePLS(n_components={self.n_components}, "
            f"forgetting_factor={self.forgetting_factor}, "
            f"scale={self.scale}, center={self.center}, "
            f"backend='{self.backend}')"
        )



"""
Trace-Based Extractor - Extract minimal pipeline from execution trace.

This module provides the TraceBasedExtractor class which extracts the minimal
set of pipeline steps needed to replay a prediction from an ExecutionTrace.

The extractor is controller-agnostic: it uses the recorded trace information
to identify which steps are needed, without encoding knowledge of specific
controller types. This enables the same infrastructure to work with any
controller (existing or custom).

Key Features:
    - Extracts minimal pipeline (only steps needed for prediction)
    - Builds artifact map for step-based artifact injection
    - Creates dependency graph for step execution order
    - Supports branch-aware extraction for multi-branch pipelines

Usage:
    >>> from nirs4all.pipeline.trace import TraceBasedExtractor, ExecutionTrace
    >>>
    >>> # Load trace from manifest
    >>> trace = ExecutionTrace.from_dict(trace_dict)
    >>>
    >>> # Extract minimal pipeline
    >>> extractor = TraceBasedExtractor()
    >>> minimal = extractor.extract(trace, full_pipeline_steps)
    >>>
    >>> # Use minimal pipeline for prediction
    >>> for step_info in minimal.steps:
    ...     artifacts = minimal.get_artifacts_for_step(step_info.step_index)
    ...     # Execute step with artifacts
"""

import logging
from dataclasses import dataclass, field
from typing import Any, Optional

from nirs4all.pipeline.trace.execution_trace import (
    ExecutionStep,
    ExecutionTrace,
    StepArtifacts,
    StepExecutionMode,
)

logger = logging.getLogger(__name__)

@dataclass
class MinimalPipelineStep:
    """A step in the minimal pipeline for prediction replay.

    Contains the step configuration and metadata needed to replay
    the step during prediction, without encoding controller-specific logic.

    Attributes:
        step_index: 1-based step index from original pipeline
        step_config: The pipeline step configuration (dict or object)
        execution_mode: How to execute this step (train/predict/skip)
        artifacts: Artifacts for this step (from trace)
        operator_type: Type of operator (for logging/debugging)
        operator_class: Class name of operator
        branch_path: Branch path if in branch context
        branch_name: Human-readable branch name
        depends_on: Indices of steps this step depends on
    """

    step_index: int
    step_config: Any = None
    execution_mode: StepExecutionMode = StepExecutionMode.PREDICT
    artifacts: StepArtifacts = field(default_factory=StepArtifacts)
    operator_type: str = ""
    operator_class: str = ""
    branch_path: list[int] = field(default_factory=list)
    branch_name: str = ""
    substep_index: int | None = None
    depends_on: set[int] = field(default_factory=set)

    def has_artifacts(self) -> bool:
        """Check if this step has associated artifacts.

        Returns:
            True if artifacts are available for this step
        """
        return len(self.artifacts.artifact_ids) > 0

    def get_artifact_ids(self) -> list[str]:
        """Get all artifact IDs for this step.

        Returns:
            List of artifact IDs
        """
        return list(self.artifacts.artifact_ids)

    def get_artifact_by_chain(self, chain_path: str) -> str | None:
        """Get artifact ID by V3 chain path.

        Args:
            chain_path: Operator chain path

        Returns:
            Artifact ID or None if not found
        """
        return self.artifacts.get_artifact_by_chain(chain_path)

    def get_artifacts_by_chain(self) -> dict[str, str]:
        """Get all artifacts indexed by chain path.

        Returns:
            Dict mapping chain_path to artifact_id
        """
        return dict(self.artifacts.by_chain) if self.artifacts.by_chain else {}

@dataclass
class MinimalPipeline:
    """Minimal pipeline extracted from an execution trace.

    Contains only the steps needed to replay a prediction, with artifact
    mappings for each step. Used by MinimalPredictor for efficient prediction.

    Attributes:
        trace_id: ID of the source execution trace
        pipeline_uid: UID of the parent pipeline
        steps: Ordered list of minimal steps to execute
        artifact_map: Mapping of step_index to list of (artifact_id, step_artifacts)
        model_step_index: Index of the model step
        fold_weights: Per-fold weights for CV ensemble
        preprocessing_chain: Summary of preprocessing steps
        metadata: Additional metadata from trace
    """

    trace_id: str = ""
    pipeline_uid: str = ""
    steps: list[MinimalPipelineStep] = field(default_factory=list)
    artifact_map: dict[int, StepArtifacts] = field(default_factory=dict)
    model_step_index: int | None = None
    fold_weights: dict[int, float] | None = None
    preprocessing_chain: str = ""
    metadata: dict[str, Any] = field(default_factory=dict)

    def get_step(self, step_index: int) -> MinimalPipelineStep | None:
        """Get a step by its index.

        Args:
            step_index: 1-based step index

        Returns:
            MinimalPipelineStep or None if not found
        """
        for step in self.steps:
            if step.step_index == step_index:
                return step
        return None

    def get_artifacts_for_step(self, step_index: int) -> StepArtifacts | None:
        """Get artifacts for a specific step.

        Args:
            step_index: 1-based step index

        Returns:
            StepArtifacts or None if not found
        """
        return self.artifact_map.get(step_index)

    def has_step(self, step_index: int) -> bool:
        """Check if a step is included in the minimal pipeline.

        Args:
            step_index: 1-based step index

        Returns:
            True if step is included
        """
        return any(s.step_index == step_index for s in self.steps)

    def get_step_count(self) -> int:
        """Get the number of steps in the minimal pipeline.

        Returns:
            Number of steps
        """
        return len(self.steps)

    def get_artifact_ids(self) -> list[str]:
        """Get all artifact IDs in the minimal pipeline.

        Returns:
            List of all artifact IDs across all steps
        """
        artifact_ids = []
        for artifacts in self.artifact_map.values():
            artifact_ids.extend(artifacts.artifact_ids)
        return artifact_ids

    def get_step_indices(self) -> list[int]:
        """Get all step indices in execution order.

        Returns:
            List of step indices
        """
        return [s.step_index for s in self.steps]

    def get_artifact_by_chain(self, chain_path: str) -> str | None:
        """Get artifact ID by V3 chain path across all steps.

        Args:
            chain_path: Operator chain path

        Returns:
            Artifact ID or None if not found
        """
        for step_artifacts in self.artifact_map.values():
            artifact_id = step_artifacts.get_artifact_by_chain(chain_path)
            if artifact_id:
                return artifact_id
        return None

    def get_all_chain_paths(self) -> dict[str, str]:
        """Get all artifacts indexed by chain path.

        Returns:
            Dict mapping chain_path to artifact_id
        """
        chain_map = {}
        for step_artifacts in self.artifact_map.values():
            if step_artifacts.by_chain:
                chain_map.update(step_artifacts.by_chain)
        return chain_map

    def __repr__(self) -> str:
        n_steps = len(self.steps)
        n_artifacts = len(self.get_artifact_ids())
        return (
            f"MinimalPipeline(trace_id={self.trace_id!r}, "
            f"steps={n_steps}, artifacts={n_artifacts}, "
            f"model_step={self.model_step_index})"
        )

class TraceBasedExtractor:
    """Extract minimal pipeline from execution trace.

    The extractor analyzes an ExecutionTrace to determine which steps
    are needed for prediction replay and builds a MinimalPipeline with
    the correct artifact mappings.

    The extractor is controller-agnostic: it uses trace metadata
    to identify steps without encoding knowledge of controller types.

    Attributes:
        include_skipped: Whether to include skipped steps in minimal pipeline
        preserve_order: Whether to preserve original step order

    Example:
        >>> extractor = TraceBasedExtractor()
        >>> trace = ExecutionTrace.from_dict(trace_dict)
        >>> minimal = extractor.extract(trace, full_pipeline_steps)
        >>> print(f"Minimal pipeline has {minimal.get_step_count()} steps")
    """

    def __init__(
        self,
        include_skipped: bool = False,
        preserve_order: bool = True
    ):
        """Initialize trace-based extractor.

        Args:
            include_skipped: Whether to include steps that were skipped
            preserve_order: Whether to preserve original step execution order
        """
        self.include_skipped = include_skipped
        self.preserve_order = preserve_order

    def extract(
        self,
        trace: ExecutionTrace,
        full_pipeline: list[Any] | None = None,
        up_to_model: bool = True
    ) -> MinimalPipeline:
        """Extract minimal pipeline from execution trace.

        Analyzes the trace to determine which steps are needed for prediction
        and builds a MinimalPipeline with artifact mappings.

        Args:
            trace: ExecutionTrace to extract from
            full_pipeline: Optional full pipeline steps (for step configs)
            up_to_model: If True, only include steps up to model step

        Returns:
            MinimalPipeline with steps and artifact mappings
        """
        minimal = MinimalPipeline(
            trace_id=trace.trace_id,
            pipeline_uid=trace.pipeline_uid,
            model_step_index=trace.model_step_index,
            fold_weights=trace.fold_weights,
            preprocessing_chain=trace.preprocessing_chain,
            metadata=trace.metadata.copy() if trace.metadata else {}
        )

        # Get steps to include
        trace_steps = trace.get_steps_up_to_model() if up_to_model and trace.model_step_index is not None else trace.steps

        # Determine if model is in a branch (has non-empty branch_path)
        # If not (feature merge case), we need to deduplicate branch steps
        model_in_branch = False
        if trace.model_step_index is not None:
            model_step = trace.get_step(trace.model_step_index)
            if model_step and model_step.branch_path:
                model_in_branch = True

        # Track which step_indices we've already added for deduplication
        # This is needed for feature merge where branch steps should only run once
        added_step_indices: set[int] = set()

        # Build minimal steps from trace
        for exec_step in trace_steps:
            # Skip if mode is SKIP and we're not including skipped
            if exec_step.execution_mode == StepExecutionMode.SKIP and not self.include_skipped:
                continue

            # Get step config from full pipeline if available
            step_config = None
            if full_pipeline and 0 < exec_step.step_index <= len(full_pipeline):
                step_config = full_pipeline[exec_step.step_index - 1]

            # Deduplication for feature merge case:
            # When model is NOT in a branch, branch steps should only appear once
            # so the branch controller can iterate all branches internally
            is_branch_step = (
                isinstance(step_config, dict) and "branch" in step_config
            )
            if not model_in_branch and is_branch_step and exec_step.step_index in added_step_indices:
                # Skip this duplicate branch step entry, but still merge artifacts
                if exec_step.has_artifacts():
                    if exec_step.step_index in minimal.artifact_map:
                        minimal.artifact_map[exec_step.step_index].merge(exec_step.artifacts)
                    else:
                        from copy import deepcopy
                        minimal.artifact_map[exec_step.step_index] = deepcopy(exec_step.artifacts)
                continue

            # Create minimal step
            minimal_step = MinimalPipelineStep(
                step_index=exec_step.step_index,
                step_config=step_config,
                execution_mode=StepExecutionMode.PREDICT,  # Force predict mode
                artifacts=exec_step.artifacts,
                operator_type=exec_step.operator_type,
                operator_class=exec_step.operator_class,
                branch_path=list(exec_step.branch_path),
                branch_name=exec_step.branch_name
            )

            minimal.steps.append(minimal_step)
            added_step_indices.add(exec_step.step_index)

            # Add to artifact map, merging if step_index already exists
            if exec_step.has_artifacts():
                if exec_step.step_index in minimal.artifact_map:
                    minimal.artifact_map[exec_step.step_index].merge(exec_step.artifacts)
                else:
                    from copy import deepcopy
                    minimal.artifact_map[exec_step.step_index] = deepcopy(exec_step.artifacts)

        # Sort by step index if preserving order
        if self.preserve_order:
            minimal.steps.sort(key=lambda s: s.step_index)

        logger.debug(
            f"Extracted minimal pipeline: {minimal.get_step_count()} steps, "
            f"{len(minimal.get_artifact_ids())} artifacts"
        )

        return minimal

    def extract_for_step(
        self,
        trace: ExecutionTrace,
        target_step_index: int,
        full_pipeline: list[Any] | None = None
    ) -> MinimalPipeline:
        """Extract minimal pipeline up to a specific step.

        Useful for partial prediction or when targeting a specific model
        in a multi-model pipeline.

        Args:
            trace: ExecutionTrace to extract from
            target_step_index: Target step index (inclusive)
            full_pipeline: Optional full pipeline steps

        Returns:
            MinimalPipeline with steps up to target
        """
        minimal = MinimalPipeline(
            trace_id=trace.trace_id,
            pipeline_uid=trace.pipeline_uid,
            preprocessing_chain=trace.preprocessing_chain,
            metadata=trace.metadata.copy() if trace.metadata else {}
        )

        # Get steps up to target
        trace_steps = [s for s in trace.steps if s.step_index <= target_step_index]

        # Check if target is the model step
        target_step = trace.get_step(target_step_index)
        if target_step and target_step.operator_type in ("model", "meta_model"):
            minimal.model_step_index = target_step_index
            minimal.fold_weights = trace.fold_weights

        for exec_step in trace_steps:
            if exec_step.execution_mode == StepExecutionMode.SKIP and not self.include_skipped:
                continue

            step_config = None
            if full_pipeline and 0 < exec_step.step_index <= len(full_pipeline):
                step_config = full_pipeline[exec_step.step_index - 1]

            minimal_step = MinimalPipelineStep(
                step_index=exec_step.step_index,
                step_config=step_config,
                execution_mode=StepExecutionMode.PREDICT,
                artifacts=exec_step.artifacts,
                operator_type=exec_step.operator_type,
                operator_class=exec_step.operator_class,
                branch_path=list(exec_step.branch_path),
                branch_name=exec_step.branch_name
            )

            minimal.steps.append(minimal_step)

            if exec_step.has_artifacts():
                if exec_step.step_index in minimal.artifact_map:
                    minimal.artifact_map[exec_step.step_index].merge(exec_step.artifacts)
                else:
                    from copy import deepcopy
                    minimal.artifact_map[exec_step.step_index] = deepcopy(exec_step.artifacts)

        if self.preserve_order:
            minimal.steps.sort(key=lambda s: s.step_index)

        return minimal

    def extract_for_branch(
        self,
        trace: ExecutionTrace,
        branch_path: list[int],
        full_pipeline: list[Any] | None = None
    ) -> MinimalPipeline:
        """Extract minimal pipeline for a specific branch.

        Includes shared steps (before branching) plus branch-specific steps.

        Args:
            trace: ExecutionTrace to extract from
            branch_path: Branch path to extract (e.g., [0] for first branch)
            full_pipeline: Optional full pipeline steps

        Returns:
            MinimalPipeline with steps for the specified branch
        """
        minimal = MinimalPipeline(
            trace_id=trace.trace_id,
            pipeline_uid=trace.pipeline_uid,
            model_step_index=trace.model_step_index,
            fold_weights=trace.fold_weights,
            preprocessing_chain=trace.preprocessing_chain,
            metadata=trace.metadata.copy() if trace.metadata else {}
        )

        # Get steps to include (up to model or all)
        trace_steps = trace.get_steps_up_to_model() if trace.model_step_index is not None else trace.steps

        for exec_step in trace_steps:
            if exec_step.execution_mode == StepExecutionMode.SKIP and not self.include_skipped:
                continue

            # Include step if:
            # 1. It has no branch (shared/pre-branch step)
            # 2. It's on the target branch path (exact or nested within)
            include_step = False
            if not exec_step.branch_path:
                include_step = True  # Shared step
            elif exec_step.branch_path == branch_path:
                include_step = True  # Exact branch match
            elif self._is_prefix_branch(branch_path, exec_step.branch_path):
                include_step = True  # Target is prefix of step's path (nested branch)
            elif self._is_parent_branch(branch_path, exec_step.branch_path):
                include_step = True  # Step is a parent branch of target

            if not include_step:
                continue

            # Get step config - for branch substeps, extract the individual substep config
            step_config = None
            if full_pipeline and 0 < exec_step.step_index <= len(full_pipeline):
                parent_config = full_pipeline[exec_step.step_index - 1]
                # If this is a substep inside a branch, extract the specific substep config
                if exec_step.branch_path and exec_step.branch_name:
                    substep_config = self._extract_substep_config(
                        parent_config, exec_step.branch_name, exec_step.substep_index
                    )
                    if substep_config is not None:
                        step_config = substep_config
                    else:
                        # Fallback to parent config if extraction fails
                        step_config = parent_config
                else:
                    step_config = parent_config

            minimal_step = MinimalPipelineStep(
                step_index=exec_step.step_index,
                step_config=step_config,
                execution_mode=StepExecutionMode.PREDICT,
                artifacts=exec_step.artifacts,
                operator_type=exec_step.operator_type,
                operator_class=exec_step.operator_class,
                branch_path=list(exec_step.branch_path),
                branch_name=exec_step.branch_name,
                substep_index=exec_step.substep_index
            )

            minimal.steps.append(minimal_step)

            if exec_step.has_artifacts():
                # Merge artifacts for substeps with the same step_index
                if exec_step.step_index in minimal.artifact_map:
                    minimal.artifact_map[exec_step.step_index].merge(exec_step.artifacts)
                else:
                    from copy import deepcopy
                    minimal.artifact_map[exec_step.step_index] = deepcopy(exec_step.artifacts)

        if self.preserve_order:
            minimal.steps.sort(key=lambda s: s.step_index)

        return minimal

    def extract_for_branch_name(
        self,
        trace: ExecutionTrace,
        branch_name: str,
        full_pipeline: list[Any] | None = None
    ) -> MinimalPipeline:
        """Extract minimal pipeline for a specific branch by name.

        More reliable than extract_for_branch for nested branches where
        branch_id doesn't map directly to branch_path. Uses branch_name
        for matching since it's unique and stored in both predictions and trace.

        Includes shared steps (before branching) plus branch-specific steps.

        Args:
            trace: ExecutionTrace to extract from
            branch_name: Branch name to match (e.g., "branch_0_branch_0")
            full_pipeline: Optional full pipeline steps

        Returns:
            MinimalPipeline with steps for the specified branch
        """
        minimal = MinimalPipeline(
            trace_id=trace.trace_id,
            pipeline_uid=trace.pipeline_uid,
            model_step_index=trace.model_step_index,
            fold_weights=trace.fold_weights,
            preprocessing_chain=trace.preprocessing_chain,
            metadata=trace.metadata.copy() if trace.metadata else {}
        )

        # Get steps to include (up to model or all)
        trace_steps = trace.get_steps_up_to_model() if trace.model_step_index is not None else trace.steps

        for exec_step in trace_steps:
            if exec_step.execution_mode == StepExecutionMode.SKIP and not self.include_skipped:
                continue

            # Include step if:
            # 1. It has no branch (shared/pre-branch step)
            # 2. It has matching branch_name
            include_step = False
            if not exec_step.branch_path:
                include_step = True  # Shared step
            elif exec_step.branch_name == branch_name:
                include_step = True  # Exact branch name match

            if not include_step:
                continue

            # Get step config - for branch substeps, extract the individual substep config
            step_config = None
            if full_pipeline and 0 < exec_step.step_index <= len(full_pipeline):
                parent_config = full_pipeline[exec_step.step_index - 1]
                # If this is a substep inside a branch, extract the specific substep config
                if exec_step.branch_path and exec_step.branch_name:
                    substep_config = self._extract_substep_config(
                        parent_config, exec_step.branch_name, exec_step.substep_index
                    )
                    if substep_config is not None:
                        step_config = substep_config
                    else:
                        # Fallback to parent config if extraction fails
                        step_config = parent_config
                else:
                    step_config = parent_config

            minimal_step = MinimalPipelineStep(
                step_index=exec_step.step_index,
                step_config=step_config,
                execution_mode=StepExecutionMode.PREDICT,
                artifacts=exec_step.artifacts,
                operator_type=exec_step.operator_type,
                operator_class=exec_step.operator_class,
                branch_path=list(exec_step.branch_path),
                branch_name=exec_step.branch_name,
                substep_index=exec_step.substep_index
            )

            minimal.steps.append(minimal_step)

            if exec_step.has_artifacts():
                # Merge artifacts for substeps with the same step_index
                if exec_step.step_index in minimal.artifact_map:
                    minimal.artifact_map[exec_step.step_index].merge(exec_step.artifacts)
                else:
                    # Create a copy to avoid modifying the original trace
                    from copy import deepcopy
                    minimal.artifact_map[exec_step.step_index] = deepcopy(exec_step.artifacts)

        if self.preserve_order:
            minimal.steps.sort(key=lambda s: s.step_index)

        return minimal

    def get_required_artifact_ids(
        self,
        trace: ExecutionTrace,
        up_to_model: bool = True
    ) -> list[str]:
        """Get list of artifact IDs required for prediction.

        Useful for pre-loading artifacts or validating artifact availability.

        Args:
            trace: ExecutionTrace to analyze
            up_to_model: If True, only include artifacts up to model step

        Returns:
            List of artifact IDs needed for prediction
        """
        steps = trace.get_steps_up_to_model() if up_to_model and trace.model_step_index is not None else trace.steps

        artifact_ids = []
        for step in steps:
            if step.execution_mode != StepExecutionMode.SKIP:
                artifact_ids.extend(step.artifacts.artifact_ids)

        return artifact_ids

    def get_step_dependency_graph(
        self,
        trace: ExecutionTrace
    ) -> dict[int, set[int]]:
        """Build dependency graph from execution trace.

        The dependency graph maps each step to the set of steps it depends on.
        This is inferred from the trace execution order and branch structure.

        Args:
            trace: ExecutionTrace to analyze

        Returns:
            Dictionary mapping step_index to set of dependency step indices
        """
        dependencies: dict[int, set[int]] = {}
        prev_step_by_branch: dict[tuple[int, ...], int] = {}

        for step in trace.steps:
            step_idx = step.step_index
            branch_key = tuple(step.branch_path) if step.branch_path else ()
            dependencies[step_idx] = set()

            # Depend on previous step in same branch
            if branch_key in prev_step_by_branch:
                dependencies[step_idx].add(prev_step_by_branch[branch_key])

            # If in a branch, also depend on last pre-branch step
            if branch_key and () in prev_step_by_branch:
                # Find the last shared step before branching
                shared_step = prev_step_by_branch[()]
                if shared_step not in dependencies[step_idx]:
                    dependencies[step_idx].add(shared_step)

            # Update previous step for this branch
            prev_step_by_branch[branch_key] = step_idx

        return dependencies

    def _extract_substep_config(
        self,
        branch_step_config: Any,
        branch_name: str,
        substep_index: int | None
    ) -> Any:
        """Extract individual substep config from a branch step.

        When a step is inside a branch, the full_pipeline contains the branch
        step config (e.g., {'branch': {'ridge': [...], 'pls': [...]}}).
        This method extracts the specific substep config for the given branch
        and substep index.

        Args:
            branch_step_config: The branch step configuration from full_pipeline
            branch_name: Name of the branch (e.g., 'pls', 'branch_0')
            substep_index: Index of substep within the branch (0-based)

        Returns:
            The extracted substep config, or None if extraction fails
        """
        if not isinstance(branch_step_config, dict):
            return None

        if "branch" not in branch_step_config:
            return None

        branches = branch_step_config["branch"]
        if not branches:
            return None

        # Handle dict branches (named): {'branch': {'ridge': [...], 'pls': [...]}}
        if isinstance(branches, dict):
            # Try exact name match first
            if branch_name in branches:
                branch_steps = branches[branch_name]
            else:
                # For generated names like 'branch_0', try to match by index
                # E.g., 'branch_0' -> index 0 -> first branch
                branch_keys = list(branches.keys())
                if branch_name.startswith("branch_"):
                    try:
                        idx = int(branch_name.split("_")[1])
                        if 0 <= idx < len(branch_keys):
                            branch_steps = branches[branch_keys[idx]]
                        else:
                            return None
                    except (ValueError, IndexError):
                        return None
                else:
                    return None
        # Handle list branches: {'branch': [[...], [...]]}
        elif isinstance(branches, list):
            # Extract branch index from name (e.g., 'branch_0' -> 0)
            if branch_name.startswith("branch_"):
                try:
                    idx = int(branch_name.split("_")[1])
                    if 0 <= idx < len(branches):
                        branch_steps = branches[idx]
                    else:
                        return None
                except (ValueError, IndexError):
                    return None
            else:
                return None
        else:
            return None

        # Now extract substep from branch_steps
        if substep_index is not None and isinstance(branch_steps, list):
            if 0 <= substep_index < len(branch_steps):
                return branch_steps[substep_index]
            else:
                logger.warning(
                    f"Substep index {substep_index} out of range for branch "
                    f"'{branch_name}' with {len(branch_steps)} steps"
                )
                return None

        return None

    def _is_prefix_branch(
        self,
        prefix_path: list[int],
        full_path: list[int]
    ) -> bool:
        """Check if prefix_path is a prefix of full_path.

        Used to match nested branches where we want a specific outer branch
        but any inner branch. For example, [0] is a prefix of [0, 0] and [0, 1].

        Args:
            prefix_path: The potential prefix path
            full_path: The full branch path to check

        Returns:
            True if prefix_path is a proper prefix of full_path
        """
        if len(prefix_path) >= len(full_path):
            return False
        return full_path[:len(prefix_path)] == prefix_path

    def _is_parent_branch(
        self,
        target_path: list[int],
        check_path: list[int]
    ) -> bool:
        """Check if check_path is a parent of target_path.

        Args:
            target_path: Target branch path
            check_path: Path to check if it's a parent

        Returns:
            True if check_path is a prefix of target_path
        """
        if len(check_path) >= len(target_path):
            return False
        return target_path[:len(check_path)] == check_path

    def validate_trace_for_prediction(
        self,
        trace: ExecutionTrace
    ) -> tuple[bool, list[str]]:
        """Validate that a trace has all information needed for prediction.

        Checks that:
        - Model step is recorded
        - All steps up to model have recorded artifacts (if applicable)
        - No critical information is missing

        Args:
            trace: ExecutionTrace to validate

        Returns:
            Tuple of (is_valid, list of issues)
        """
        issues = []

        # Check model step
        if trace.model_step_index is None:
            issues.append("No model step recorded in trace")

        # Check steps have artifacts where expected
        steps_needing_artifacts = {"transform", "model", "y_processing",
                                   "feature_augmentation", "concat_transform",
                                   "feature_selection"}

        for step in trace.steps:
            if step.execution_mode == StepExecutionMode.SKIP:
                continue

            # Check if this type typically needs artifacts
            if step.operator_type in steps_needing_artifacts and not step.has_artifacts():
                issues.append(
                    f"Step {step.step_index} ({step.operator_type}/{step.operator_class}) "
                    f"has no recorded artifacts"
                )

        is_valid = len(issues) == 0
        return is_valid, issues



"""
PredictionCache - Caching layer for expensive prediction computations.

This module provides caching for aggregated predictions to avoid
redundant computations when multiple charts use the same data.
"""
import hashlib
import json
import time
from collections.abc import Callable
from typing import Any, Optional, Union


class CacheKey:
    """Immutable cache key for prediction queries."""

    __slots__ = ('_hash', '_repr')

    def __init__(
        self,
        aggregate: str | None,
        rank_metric: str,
        rank_partition: str,
        display_partition: str,
        group_by: tuple[str, ...] | None,
        filters: tuple[tuple[str, Any], ...]
    ):
        """Create a cache key from query parameters.

        Args:
            aggregate: Aggregation column name (e.g., 'ID') or None.
            rank_metric: Metric used for ranking.
            rank_partition: Partition used for ranking.
            display_partition: Partition for display data.
            group_by: Tuple of column names for grouping.
            filters: Tuple of (key, value) filter pairs.
        """
        # Create a hashable representation
        self._repr = (
            aggregate,
            rank_metric,
            rank_partition,
            display_partition,
            group_by,
            filters
        )
        self._hash = hash(self._repr)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, CacheKey):
            return False
        return self._repr == other._repr

    def __repr__(self):
        return f"CacheKey({self._repr})"

class PredictionCache:
    """LRU-style cache for prediction query results.

    Caches expensive computations like aggregations to speed up
    repeated chart rendering with the same parameters.

    The cache stores:
    - Aggregated prediction results (keyed by aggregate + filters)
    - Top-k results (keyed by full query parameters)

    Cache invalidation happens when:
    - Max size is exceeded (LRU eviction)
    - clear() is called explicitly
    - Predictions data is modified (user must call clear())

    Example:
        >>> cache = PredictionCache(max_entries=100)
        >>> # First call computes and caches
        >>> result = cache.get_or_compute(key, compute_fn)
        >>> # Second call returns cached result
        >>> result = cache.get_or_compute(key, compute_fn)  # Fast!
    """

    def __init__(self, max_entries: int = 50):
        """Initialize the cache.

        Args:
            max_entries: Maximum number of cached entries before LRU eviction.
        """
        self._cache: dict[CacheKey, Any] = {}
        self._access_order: list[CacheKey] = []
        self._max_entries = max_entries
        self._stats = {
            'hits': 0,
            'misses': 0,
            'evictions': 0,
            'total_compute_time': 0.0
        }

    @staticmethod
    def make_key(
        aggregate: str | None,
        rank_metric: str,
        rank_partition: str = 'val',
        display_partition: str = 'test',
        group_by: str | list[str] | None = None,
        **filters
    ) -> CacheKey:
        """Create a cache key from query parameters.

        Args:
            aggregate: Aggregation column name or None.
            rank_metric: Metric for ranking.
            rank_partition: Partition for ranking.
            display_partition: Partition for display.
            group_by: Grouping column(s).
            **filters: Additional filter criteria.

        Returns:
            CacheKey for the query.
        """
        # Normalize group_by to tuple
        group_by_tuple: tuple[str, ...] | None
        if group_by is None:
            group_by_tuple = None
        elif isinstance(group_by, str):
            group_by_tuple = (group_by,)
        else:
            group_by_tuple = tuple(sorted(group_by))

        # Normalize filters to sorted tuple of tuples
        # Convert values to strings for hashability
        filter_items = []
        for k, v in sorted(filters.items()):
            if v is None:
                continue
            # Convert lists to tuples for hashability
            if isinstance(v, list):
                v = tuple(v)
            filter_items.append((k, v))

        return CacheKey(
            aggregate=aggregate,
            rank_metric=rank_metric,
            rank_partition=rank_partition,
            display_partition=display_partition,
            group_by=group_by_tuple,
            filters=tuple(filter_items)
        )

    def get(self, key: CacheKey) -> Any | None:
        """Get cached result if available.

        Args:
            key: Cache key to look up.

        Returns:
            Cached result or None if not found.
        """
        if key in self._cache:
            self._stats['hits'] += 1
            # Update access order (move to end)
            if key in self._access_order:
                self._access_order.remove(key)
            self._access_order.append(key)
            return self._cache[key]

        self._stats['misses'] += 1
        return None

    def put(self, key: CacheKey, value: Any) -> None:
        """Store result in cache.

        Args:
            key: Cache key.
            value: Result to cache.
        """
        # Evict oldest entries if at capacity
        while len(self._cache) >= self._max_entries:
            self._evict_oldest()

        self._cache[key] = value
        if key in self._access_order:
            self._access_order.remove(key)
        self._access_order.append(key)

    def get_or_compute(self, key: CacheKey, compute_fn: Callable) -> Any:
        """Get cached result or compute and cache.

        This is the primary method for cache usage. It handles
        the cache lookup, computation, and storage in one call.

        Args:
            key: Cache key.
            compute_fn: Function to call if cache miss (no args).

        Returns:
            Result from cache or computation.
        """
        result = self.get(key)
        if result is not None:
            return result

        # Compute and cache
        t0 = time.time()
        result = compute_fn()
        compute_time = time.time() - t0
        self._stats['total_compute_time'] += compute_time

        self.put(key, result)
        return result

    def _evict_oldest(self) -> None:
        """Evict the least recently used entry."""
        if self._access_order:
            oldest_key = self._access_order.pop(0)
            if oldest_key in self._cache:
                del self._cache[oldest_key]
                self._stats['evictions'] += 1

    def clear(self) -> None:
        """Clear all cached entries."""
        self._cache.clear()
        self._access_order.clear()

    def get_stats(self) -> dict[str, Any]:
        """Get cache statistics.

        Returns:
            Dictionary with hits, misses, evictions, hit_rate, etc.
        """
        total = self._stats['hits'] + self._stats['misses']
        hit_rate = self._stats['hits'] / total if total > 0 else 0.0

        return {
            **self._stats,
            'total_requests': total,
            'hit_rate': hit_rate,
            'current_size': len(self._cache),
            'max_size': self._max_entries
        }

    def __len__(self) -> int:
        """Return number of cached entries."""
        return len(self._cache)

    def __repr__(self) -> str:
        stats = self.get_stats()
        return f"PredictionCache({stats['current_size']}/{stats['max_size']} entries, {stats['hit_rate']:.1%} hit rate)"



"""POP-PLS: Per-Operator-Per-component PLS regressor for nirs4all.

A sklearn-compatible implementation of POP-PLS that selects a potentially
DIFFERENT linear preprocessing operator for EACH PLS component. This is the
true multi-operator mode: each component independently chooses the operator
that best improves the model, then extracts that component through the
selected operator's lens via NIPALS deflation.

Compared to AOM-PLS (which selects ONE operator for ALL components), POP-PLS
can adapt its preprocessing strategy as successive residuals reveal different
spectral features, e.g. component 1 might use SG 1st derivative to capture
slope information while component 3 uses detrending for baseline correction.

Mathematical formulation
------------------------
Let X in R^{n x p} be the input matrix and y in R^n the response vector.
Given an operator bank {A_b}_{b=1..B} of p x p linear operators:

For k = 1..K_max:
  1. Cross-covariance: c_k = X_res^T @ y_res
  2. For each operator b in the bank:
     - Adjoint trick: g_b = A_b^T @ c_k
     - Normalize: w_hat_b = g_b / ||g_b||
     - Forward pass: a_w_b = A_b @ w_hat_b, then w_b = a_w_b / ||a_w_b||
     - Candidate score: t_b = X_res @ w_b
     - Build prefix model B_k, evaluate PRESS on full training data
  3. Select: b* = argmin_b PRESS(prefix_k)
  4. Extract component k using operator b*: NIPALS deflation
  5. Record gamma_[k, b*] = 1.0

Built-in model selection via PRESS (no Optuna or holdout needed):
  PRESS (Predicted Residual Error Sum of Squares) approximates leave-one-out
  cross-validation using the PLS hat matrix: PRESS = sum((e_i / (1 - h_ii))^2).
  This uses ALL training data for selection (no holdout split), providing
  stable, deterministic operator and component-count decisions. The n_orth
  OPLS pre-filter is auto-tuned by comparing PRESS across candidates.

References
----------
- de Jong, S. (1993). SIMPLS: An alternative approach to partial least
  squares regression. Chemometrics and Intelligent Laboratory Systems.
- Martens, M. & Martens, H. (2001). Multivariate Analysis of Quality:
  An Introduction. Wiley.
- Allen, D.M. (1974). The Relationship Between Variable Selection and
  Data Augmentation and a Method for Prediction. Technometrics.
"""

from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike, NDArray
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted

from nirs4all.operators.models.sklearn.aom_pls import (
    ComposedOperator,
    DetrendProjectionOperator,
    IdentityOperator,
    LinearOperator,
    SavitzkyGolayOperator,
    _opls_prefilter,
)

# =============================================================================
# POP-PLS Operator Bank
# =============================================================================

def pop_pls_operator_bank() -> list[LinearOperator]:
    """Build a dedicated operator bank optimized for POP-PLS.

    Unlike the AOM bank (which has many parametric variants for global search),
    this bank is COMPACT and DIVERSE: ~9 operators, each capturing a
    qualitatively different spectral transformation. This design exploits
    POP-PLS's per-component selection: instead of many similar operators
    competing (adding noise to selection), a few diverse operators each serve
    a clear role at different stages of NIPALS deflation.

    Design principles:
    - Maximum diversity: each operator has a qualitatively different effect
    - Spanning mild to aggressive: from identity to second derivatives
    - Covering the key NIRS preprocessing families: smoothing, derivatives,
      detrending, and compositions thereof
    - Compact (9 operators) to minimize selection noise on small holdouts
      or PRESS evaluations

    Returns
    -------
    operators : list of LinearOperator
        Compact, diverse operator bank for POP-PLS (~9 operators).
    """
    return [
        # Identity — recovers standard PLS, always a contender for early
        # components where dominant features need no preprocessing.
        IdentityOperator(),
        # Smoothing — reduces high-frequency noise while preserving spectral
        # shape. Good for early components where dominant variation is broad.
        SavitzkyGolayOperator(window=21, polyorder=2, deriv=0),
        # 1st derivative, narrow window — enhances sharp features, removes
        # baseline. The workhorse of NIRS preprocessing. High spectral
        # resolution but amplifies noise.
        SavitzkyGolayOperator(window=11, polyorder=2, deriv=1),
        # 1st derivative, medium window — broader feature enhancement with
        # better noise suppression. Balances resolution and stability.
        SavitzkyGolayOperator(window=21, polyorder=2, deriv=1),
        # 1st derivative, wide window — maximum noise suppression derivative.
        # Good for late components where residuals are noisy.
        SavitzkyGolayOperator(window=31, polyorder=2, deriv=1),
        # 2nd derivative — removes both baseline and linear trends. Isolates
        # peak curvature. Good for overlapping spectral bands.
        SavitzkyGolayOperator(window=21, polyorder=3, deriv=2),
        # Linear detrend — removes linear baseline drift without computing
        # derivatives. Complementary to derivative approaches.
        DetrendProjectionOperator(degree=1),
        # Quadratic detrend — removes curved baseline caused by scattering
        # effects. Handles more complex baselines than linear detrend.
        DetrendProjectionOperator(degree=2),
        # Combined 1st derivative + detrend — maximum baseline removal with
        # feature enhancement. Good for complex baseline scenarios where
        # neither derivative nor detrend alone suffices.
        ComposedOperator(
            first=SavitzkyGolayOperator(window=15, polyorder=2, deriv=1),
            second=DetrendProjectionOperator(degree=1),
        ),
    ]

# =============================================================================
# POP-PLS Core Algorithms
# =============================================================================

def _compute_weight(op: LinearOperator, c_k: NDArray, eps: float) -> NDArray | None:
    """Compute PLS weight vector via the adjoint trick for a given operator.

    Returns None if the operator produces degenerate results.
    """
    g_b = op.apply_adjoint(c_k)
    g_norm = np.linalg.norm(g_b)
    if g_norm < eps:
        return None
    w_hat_b = g_b / g_norm
    a_w = op.apply(w_hat_b.reshape(1, -1)).ravel()
    a_w_norm = np.linalg.norm(a_w)
    if a_w_norm < eps:
        return None
    return np.asarray(a_w / a_w_norm)

def _compute_prefix_B(W: NDArray, P: NDArray, Q: NDArray, k: int) -> NDArray:
    """Compute prefix regression coefficient matrix B_k = W @ inv(P^T W) @ Q^T."""
    PtW = P[:, :k].T @ W[:, :k]
    try:
        R_k = W[:, :k] @ np.linalg.inv(PtW)
    except np.linalg.LinAlgError:
        R_k = W[:, :k] @ np.linalg.pinv(PtW)
    return np.asarray(R_k @ Q[:, :k].T)

def _poppls_fit_greedy(
    X: NDArray, Y: NDArray, operators: list[LinearOperator],
    n_components: int, n_orth: int, eps: float = 1e-12,
) -> dict:
    """Greedy POP-PLS extraction using training R^2 criterion.

    Used when auto_select=False. Each component greedily selects the operator
    with highest R^2(y_res, t_b) on training data. All extractable components
    are returned without model selection.
    """
    n, p = X.shape
    q = Y.shape[1]
    B = len(operators)

    P_orth = None
    if n_orth > 0:
        X, P_orth, _ = _opls_prefilter(X, Y[:, 0], n_orth)

    W = np.zeros((p, n_components), dtype=np.float64)
    T = np.zeros((n, n_components), dtype=np.float64)
    P = np.zeros((p, n_components), dtype=np.float64)
    Q = np.zeros((q, n_components), dtype=np.float64)
    Gamma = np.zeros((n_components, B), dtype=np.float64)
    component_operators = []

    X_res = X.copy()
    Y_res = Y.copy()
    n_extracted = 0

    for k in range(n_components):
        c_k = X_res.T @ Y_res
        if q == 1:
            c_k = c_k[:, 0]
            y_score = Y_res[:, 0]
        else:
            u, s, vt = np.linalg.svd(c_k, full_matrices=False)
            c_k = u[:, 0] * s[0]
            y_score = Y_res @ vt[0]

        c_norm = np.linalg.norm(c_k)
        if c_norm < eps:
            break

        y_norm_sq = np.dot(y_score, y_score)
        if y_norm_sq < eps:
            break

        best_r2 = -np.inf
        best_b = 0
        best_w = None
        best_t = None

        for b, op in enumerate(operators):
            w_b = _compute_weight(op, c_k, eps)
            if w_b is None:
                continue
            t_b = X_res @ w_b
            t_norm_sq = np.dot(t_b, t_b)
            if t_norm_sq < eps:
                continue
            cov_yt = np.dot(y_score, t_b)
            r2 = cov_yt ** 2 / (y_norm_sq * t_norm_sq)
            if r2 > best_r2:
                best_r2 = r2
                best_b = b
                best_w = w_b
                best_t = t_b

        if best_w is None or best_t is None:
            break

        t_k = best_t
        tt = t_k @ t_k
        if tt < eps:
            break

        p_k = (X_res.T @ t_k) / tt
        q_k = (Y_res.T @ t_k) / tt

        W[:, k] = best_w
        T[:, k] = t_k
        P[:, k] = p_k
        Q[:, k] = q_k
        Gamma[k, best_b] = 1.0
        component_operators.append(operators[best_b].name)
        n_extracted = k + 1

        X_res -= np.outer(t_k, p_k)
        Y_res -= np.outer(t_k, q_k)

    W = W[:, :n_extracted]
    T = T[:, :n_extracted]
    P = P[:, :n_extracted]
    Q = Q[:, :n_extracted]
    Gamma = Gamma[:n_extracted]

    B_coefs = np.zeros((n_extracted, p, q), dtype=np.float64)
    for k in range(n_extracted):
        B_coefs[k] = _compute_prefix_B(W, P, Q, k + 1)

    return {
        "n_extracted": n_extracted, "k_selected": n_extracted,
        "W": W, "T": T, "P": P, "Q": Q, "Gamma": Gamma,
        "B_coefs": B_coefs, "P_orth": P_orth,
        "component_operators": component_operators,
    }

def _poppls_holdout_pass(
    X_train: NDArray, Y_train: NDArray, operators: list[LinearOperator],
    n_components: int, n_orth: int, X_val: NDArray, Y_val: NDArray,
    eps: float = 1e-12,
) -> dict | None:
    """Single holdout pass for external validation data.

    At each component, selects the operator that minimizes the prefix RMSE on
    the external validation set. Tracks the best prefix count across all
    components. Uses patience-based early stopping.

    Returns None if no components could be extracted.
    """
    n, p = X_train.shape
    q = Y_train.shape[1]
    B = len(operators)

    # OPLS pre-filter
    P_orth = None
    X_t = X_train.copy()
    X_v = X_val.copy()
    if n_orth > 0:
        X_t, P_orth, _ = _opls_prefilter(X_t, Y_train[:, 0], n_orth)
        for j in range(P_orth.shape[1]):
            p_o = P_orth[:, j]
            t_o = X_v @ p_o
            X_v = X_v - np.outer(t_o, p_o)

    max_comp = min(n - 1, p, n_components)
    if max_comp <= 0:
        return None

    W = np.zeros((p, max_comp), dtype=np.float64)
    P = np.zeros((p, max_comp), dtype=np.float64)
    Q = np.zeros((q, max_comp), dtype=np.float64)
    operator_indices: list[int] = []

    X_res = X_t.copy()
    Y_res = Y_train.copy()
    n_extracted = 0

    # Baseline: no components (predict 0 in centered space)
    baseline_rmse = np.sqrt(np.mean(Y_val ** 2))
    best_rmse = baseline_rmse
    best_k = 0
    patience = 5
    no_improve = 0

    for k in range(max_comp):
        c_k = X_res.T @ Y_res
        if q == 1:
            c_k = c_k[:, 0]
        else:
            u, s, vt = np.linalg.svd(c_k, full_matrices=False)
            c_k = u[:, 0] * s[0]

        c_norm = np.linalg.norm(c_k)
        if c_norm < eps:
            break

        # Evaluate each operator using holdout prefix RMSE
        best_op_rmse = np.inf
        best_op_b = None
        best_op_artifacts = None

        for b, op in enumerate(operators):
            w_b = _compute_weight(op, c_k, eps)
            if w_b is None:
                continue
            t_b = X_res @ w_b
            tt = np.dot(t_b, t_b)
            if tt < eps:
                continue

            p_b = (X_res.T @ t_b) / tt
            q_b = (Y_res.T @ t_b) / tt

            # Build temporary prefix model and evaluate on holdout
            W_tmp = np.column_stack([W[:, :k], w_b]) if k > 0 else w_b.reshape(-1, 1)
            P_tmp = np.column_stack([P[:, :k], p_b]) if k > 0 else p_b.reshape(-1, 1)
            Q_tmp = np.column_stack([Q[:, :k], q_b]) if k > 0 else q_b.reshape(-1, 1)

            B_k = _compute_prefix_B(W_tmp, P_tmp, Q_tmp, k + 1)
            y_pred_val = X_v @ B_k
            rmse = np.sqrt(np.mean((Y_val - y_pred_val) ** 2))

            if rmse < best_op_rmse:
                best_op_rmse = rmse
                best_op_b = b
                best_op_artifacts = (w_b, t_b, p_b, q_b)

        if best_op_b is None or best_op_artifacts is None:
            break

        w_k, t_k, p_k, q_k = best_op_artifacts

        W[:, k] = w_k
        P[:, k] = p_k
        Q[:, k] = q_k
        operator_indices.append(best_op_b)
        n_extracted = k + 1

        # NIPALS deflation
        X_res -= np.outer(t_k, p_k)
        Y_res -= np.outer(t_k, q_k)

        # Track best prefix
        if best_op_rmse < best_rmse:
            best_rmse = best_op_rmse
            best_k = n_extracted
            no_improve = 0
        else:
            no_improve += 1
            if no_improve >= patience:
                break

    if n_extracted == 0:
        return None

    return {
        "n_orth": n_orth,
        "k_selected": best_k,
        "operator_indices": operator_indices,
        "best_rmse": best_rmse,
    }

def _poppls_press_pass(
    X: NDArray, Y: NDArray, operators: list[LinearOperator],
    n_components: int, n_orth: int, eps: float = 1e-12,
) -> dict | None:
    """PRESS-based POP-PLS: operator and component selection via LOO-CV.

    Uses the PRESS (Predicted Residual Error Sum of Squares) statistic for
    per-component operator selection and component count determination. No
    holdout split is needed — PRESS approximates leave-one-out cross-validation
    from the full training data using the PLS hat matrix leverages:

        PRESS_k = sum_i (e_i / (1 - h_ii))^2

    where h_ii = sum_j t_{ij}^2 / ||t_j||^2 are the leverage values from the
    orthogonal PLS score matrix.

    Since PRESS uses ALL training data, it provides more stable operator and
    component selections than holdout-based approaches, especially for small
    datasets. The selection is also deterministic (no random splits).

    Returns a complete model dictionary (no refit phase needed since all data
    is used during selection).
    """
    n, p = X.shape
    q = Y.shape[1]
    B_ops = len(operators)

    # OPLS pre-filter
    P_orth = None
    if n_orth > 0:
        X, P_orth, _ = _opls_prefilter(X, Y[:, 0], n_orth)

    # Limit components: n-2 to ensure leverages stay < 1
    max_comp = min(n - 2, p, n_components)
    if max_comp <= 0:
        return None

    W = np.zeros((p, max_comp), dtype=np.float64)
    T_scores = np.zeros((n, max_comp), dtype=np.float64)
    P_load = np.zeros((p, max_comp), dtype=np.float64)
    Q_load = np.zeros((q, max_comp), dtype=np.float64)
    Gamma = np.zeros((max_comp, B_ops), dtype=np.float64)
    operator_indices: list[int] = []
    component_operators: list[str] = []

    # Store X before NIPALS deflation (needed for prefix predictions)
    X_orig = X.copy()

    X_res = X.copy()
    Y_res = Y.copy()
    n_extracted = 0

    # Baseline PRESS: predicting zero (centered data). For LOO with mean-only
    # model: e_{-i} = y_i * n/(n-1), so PRESS_0 = (n/(n-1))^2 * ||Y||_F^2
    baseline_press = (n / (n - 1)) ** 2 * np.sum(Y ** 2)
    best_press = baseline_press
    best_k = 0

    # Cumulative leverages from extracted components.
    # Since NIPALS scores are orthogonal: h_ii = sum_j t_{ij}^2 / ||t_j||^2
    h_cumulative = np.zeros(n, dtype=np.float64)

    patience = 5
    no_improve = 0

    for k in range(max_comp):
        c_k = X_res.T @ Y_res
        if q == 1:
            c_k = c_k[:, 0]
        else:
            u, s, vt = np.linalg.svd(c_k, full_matrices=False)
            c_k = u[:, 0] * s[0]

        if np.linalg.norm(c_k) < eps:
            break

        best_op_press = np.inf
        best_op_b = None
        best_op_artifacts = None
        best_op_h_increment = None

        for b, op in enumerate(operators):
            w_b = _compute_weight(op, c_k, eps)
            if w_b is None:
                continue
            t_b = X_res @ w_b
            tt = np.dot(t_b, t_b)
            if tt < eps:
                continue

            p_b = (X_res.T @ t_b) / tt
            q_b = (Y_res.T @ t_b) / tt

            # Build temporary prefix model
            W_tmp = np.column_stack([W[:, :k], w_b]) if k > 0 else w_b.reshape(-1, 1)
            P_tmp = np.column_stack([P_load[:, :k], p_b]) if k > 0 else p_b.reshape(-1, 1)
            Q_tmp = np.column_stack([Q_load[:, :k], q_b]) if k > 0 else q_b.reshape(-1, 1)

            # Prefix prediction on original (pre-deflation) X
            B_k = _compute_prefix_B(W_tmp, P_tmp, Q_tmp, k + 1)
            y_pred = X_orig @ B_k
            residuals = Y - y_pred

            # Leverage increment from candidate component
            h_increment = t_b ** 2 / tt
            h_total = h_cumulative + h_increment

            # Clip leverages for numerical stability (avoid division by ~0)
            h_clipped = np.minimum(h_total, 1 - 1e-6)

            # PRESS: sum_i (e_i / (1 - h_ii))^2
            press = np.sum((residuals / (1 - h_clipped[:, np.newaxis])) ** 2)

            if press < best_op_press:
                best_op_press = press
                best_op_b = b
                best_op_artifacts = (w_b, t_b, p_b, q_b)
                best_op_h_increment = h_increment

        if best_op_b is None or best_op_artifacts is None or best_op_h_increment is None:
            break

        # Accept this component
        w_k, t_k, p_k, q_k = best_op_artifacts
        W[:, k] = w_k
        T_scores[:, k] = t_k
        P_load[:, k] = p_k
        Q_load[:, k] = q_k
        Gamma[k, best_op_b] = 1.0
        operator_indices.append(best_op_b)
        component_operators.append(operators[best_op_b].name)
        n_extracted = k + 1
        h_cumulative += best_op_h_increment

        # NIPALS deflation
        X_res -= np.outer(t_k, p_k)
        Y_res -= np.outer(t_k, q_k)

        # Track best PRESS
        if best_op_press < best_press:
            best_press = best_op_press
            best_k = n_extracted
            no_improve = 0
        else:
            no_improve += 1
            if no_improve >= patience:
                break

    if n_extracted == 0:
        return None

    # Trim to extracted components
    W = W[:, :n_extracted]
    T_scores = T_scores[:, :n_extracted]
    P_load = P_load[:, :n_extracted]
    Q_load = Q_load[:, :n_extracted]
    Gamma = Gamma[:n_extracted]

    # Compute all prefix B coefficients
    B_coefs = np.zeros((n_extracted, p, q), dtype=np.float64)
    for k_idx in range(n_extracted):
        B_coefs[k_idx] = _compute_prefix_B(W, P_load, Q_load, k_idx + 1)

    return {
        "n_extracted": n_extracted,
        "k_selected": best_k,
        "W": W, "T": T_scores, "P": P_load, "Q": Q_load,
        "Gamma": Gamma, "B_coefs": B_coefs, "P_orth": P_orth,
        "component_operators": component_operators,
        "best_press": best_press,
    }

def _poppls_refit(
    X: NDArray, Y: NDArray, operators: list[LinearOperator],
    config: dict, eps: float = 1e-12,
) -> dict:
    """Refit on full data using the determined operator sequence."""
    n, p = X.shape
    q = Y.shape[1]
    B = len(operators)

    n_orth = config["n_orth"]
    k_selected = config["k_selected"]
    op_indices = config["operator_indices"]

    # OPLS pre-filter on full data
    P_orth = None
    if n_orth > 0:
        X, P_orth, _ = _opls_prefilter(X, Y[:, 0], n_orth)

    W = np.zeros((p, k_selected), dtype=np.float64)
    T = np.zeros((n, k_selected), dtype=np.float64)
    P = np.zeros((p, k_selected), dtype=np.float64)
    Q = np.zeros((q, k_selected), dtype=np.float64)
    Gamma = np.zeros((k_selected, B), dtype=np.float64)
    component_operators: list[str] = []

    X_res = X.copy()
    Y_res = Y.copy()
    n_extracted = 0

    for k in range(k_selected):
        op_idx = op_indices[k]
        op = operators[op_idx]

        c_k = X_res.T @ Y_res
        if q == 1:
            c_k = c_k[:, 0]
        else:
            u, s, vt = np.linalg.svd(c_k, full_matrices=False)
            c_k = u[:, 0] * s[0]

        c_norm = np.linalg.norm(c_k)
        if c_norm < eps:
            break

        w_k = _compute_weight(op, c_k, eps)
        if w_k is None:
            break

        t_k = X_res @ w_k
        tt = np.dot(t_k, t_k)
        if tt < eps:
            break

        p_k = (X_res.T @ t_k) / tt
        q_k = (Y_res.T @ t_k) / tt

        W[:, k] = w_k
        T[:, k] = t_k
        P[:, k] = p_k
        Q[:, k] = q_k
        Gamma[k, op_idx] = 1.0
        component_operators.append(op.name)
        n_extracted = k + 1

        X_res -= np.outer(t_k, p_k)
        Y_res -= np.outer(t_k, q_k)

    # Trim
    W = W[:, :n_extracted]
    T = T[:, :n_extracted]
    P = P[:, :n_extracted]
    Q = Q[:, :n_extracted]
    Gamma = Gamma[:n_extracted]

    B_coefs = np.zeros((n_extracted, p, q), dtype=np.float64)
    for k in range(n_extracted):
        B_coefs[k] = _compute_prefix_B(W, P, Q, k + 1)

    return {
        "n_extracted": n_extracted, "k_selected": n_extracted,
        "W": W, "T": T, "P": P, "Q": Q, "Gamma": Gamma,
        "B_coefs": B_coefs, "P_orth": P_orth,
        "component_operators": component_operators,
    }

def _poppls_fit_numpy(
    X: NDArray, Y: NDArray, operators: list[LinearOperator],
    n_components: int, n_orth: int, auto_select: bool,
    X_val: NDArray | None = None, Y_val: NDArray | None = None,
    random_state: int | None = None,
) -> dict:
    """Fit POP-PLS model with automatic model selection.

    Selection strategy depends on available data:

    1. **auto_select=False**: Greedy extraction with R^2 criterion, no model
       selection. Uses all data directly.

    2. **auto_select=True, external validation provided**: Per-component
       holdout selection using the external validation set, with n_orth
       auto-tuning. Refit on full data with the selected configuration.

    3. **auto_select=True, no external validation** (default): PRESS-based
       selection using the full training set. PRESS (Predicted Residual Error
       Sum of Squares) approximates leave-one-out CV via the PLS hat matrix,
       providing stable, deterministic selection without any holdout split.
       No refit phase needed since PRESS uses all data.
    """
    if not auto_select:
        return _poppls_fit_greedy(X, Y, operators, n_components, n_orth)

    n, p = X.shape
    q = Y.shape[1]
    B = len(operators)
    eps = 1e-12

    has_external_val = X_val is not None and Y_val is not None and X_val.shape[0] > 0

    # Auto-tune n_orth: search [0..5] when n_orth=0, [0..n_orth] otherwise
    n_orth_candidates = list(range(n_orth + 1)) if n_orth > 0 else [0, 1, 2, 3, 4, 5]

    if has_external_val:
        assert X_val is not None and Y_val is not None
        # External validation path: holdout selection + refit.
        # Run per-component selection using external validation data for
        # each n_orth candidate, then refit best config on full data.
        best_cfg = None
        best_rmse = np.inf
        for n_orth_cand in n_orth_candidates:
            cfg = _poppls_holdout_pass(
                X, Y, operators, n_components, n_orth_cand,
                X_val, Y_val, eps,
            )
            if cfg is not None and cfg["best_rmse"] < best_rmse:
                best_rmse = cfg["best_rmse"]
                best_cfg = cfg

        if best_cfg is None or best_cfg["k_selected"] == 0:
            return {
                "n_extracted": 0, "k_selected": 0,
                "W": np.empty((p, 0)), "T": np.empty((n, 0)),
                "P": np.empty((p, 0)), "Q": np.empty((q, 0)),
                "Gamma": np.empty((0, B)), "B_coefs": np.empty((0, p, q)),
                "P_orth": None, "component_operators": [],
            }

        return _poppls_refit(X, Y, operators, best_cfg, eps)

    else:
        # Internal selection: PRESS-based (no holdout split needed).
        # For each n_orth candidate, run full PRESS-based POP-PLS on all
        # data. Select the n_orth with lowest PRESS. The resulting model
        # IS the final model (no refit needed).
        candidates = []
        for n_orth_cand in n_orth_candidates:
            result = _poppls_press_pass(
                X, Y, operators, n_components, n_orth_cand, eps,
            )
            if result is not None and result["k_selected"] > 0:
                candidates.append(result)

        if candidates:
            best_model = min(candidates, key=lambda r: r["best_press"])
            return best_model

        # No components improve over baseline — return empty model
        return {
            "n_extracted": 0, "k_selected": 0,
            "W": np.empty((p, 0)), "T": np.empty((n, 0)),
            "P": np.empty((p, 0)), "Q": np.empty((q, 0)),
            "Gamma": np.empty((0, B)), "B_coefs": np.empty((0, p, q)),
            "P_orth": None, "component_operators": [],
        }

# =============================================================================
# POPPLSRegressor
# =============================================================================

class POPPLSRegressor(BaseEstimator, RegressorMixin):
    """Per-Operator-Per-component PLS regressor.

    Selects a potentially DIFFERENT preprocessing operator for EACH PLS
    component, choosing the operator that best reduces prediction error.
    This is the true multi-operator mode: each component independently
    adapts its preprocessing to the spectral features visible in the residual
    at that stage of deflation.

    Includes built-in model selection via PRESS (Predicted Residual Error
    Sum of Squares) or external validation data, with automatic n_orth
    tuning, eliminating the need for external hyperparameter tuning.

    Parameters
    ----------
    n_components : int, default=27
        Maximum number of PLS components to extract. When auto_select=True,
        the actual number used for prediction is determined automatically
        by PRESS minimization.
    operator_bank : list of LinearOperator or None, default=None
        Explicit list of operators. If None, uses pop_pls_operator_bank()
        which provides a compact bank of ~9 diverse operators optimized
        for per-component selection.
    n_orth : int, default=0
        OPLS orthogonal pre-filter control. When auto_select=True and n_orth=0,
        the algorithm auto-searches over [0, 1, 2, 3, 4, 5]. When n_orth > 0,
        it searches [0, 1, ..., n_orth]. When auto_select=False, n_orth is used
        directly without search.
    center : bool, default=True
        Whether to center X and Y (subtract mean).
    scale : bool, default=False
        Whether to scale X and Y to unit variance per column.
        WARNING: per-column scaling destroys spectral shape and cripples
        SG/detrend operators. Only enable if your data is not spectral.
    auto_select : bool, default=True
        If True, automatically select operator sequence, component count,
        and n_orth. Uses PRESS (LOO-CV approximation) when no external
        validation is provided, or holdout RMSE with external validation.
        If False, uses greedy R^2 criterion without model selection.
    random_state : int or None, default=None
        Random state (currently unused since PRESS is deterministic, but
        kept for API compatibility).

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during fit.
    n_components_ : int
        Actual number of components extracted.
    k_selected_ : int
        Selected number of components for prediction.
    gamma_ : ndarray of shape (n_components_, n_blocks)
        Per-component operator selection matrix. Each row has exactly one
        1.0 entry indicating which operator was selected for that component.
    coef_ : ndarray of shape (n_features, n_targets)
        Regression coefficients using selected components.
    block_names_ : list of str
        Names of operators in the bank.
    component_operators_ : list of str
        Operator name selected for each component (length n_components_).
    selected_n_orth_ : int
        The n_orth value selected by auto-tuning (only set when auto_select=True).

    Notes
    -----
    **Why POP-PLS does not need Optuna (unlike AOM-PLS)**

    AOM-PLS selects ONE operator for ALL components. Since the optimal
    operator is unknown a priori, AOM-PLS relies on external search
    (Optuna, grid search) to find the globally best operator.

    POP-PLS eliminates this external search by making all decisions
    internally:

    1. **Operator selection** is per-component: at each deflation step, the
       operator minimizing PRESS is chosen. PRESS approximates leave-one-out
       CV using the PLS hat matrix leverages, providing stable evaluation
       on the FULL training set (no holdout variance).

    2. **Component count** is selected by tracking PRESS across prefixes.
       Early stopping with patience prevents extracting unnecessary components.

    3. **n_orth** is auto-tuned by comparing PRESS across candidates
       [0, 1, 2, 3, 4, 5].

    4. **No refit phase** is needed: since PRESS uses all training data,
       the model built during selection IS the final model.

    5. **Dedicated operator bank**: pop_pls_operator_bank() provides ~9
       diverse operators (identity, smoothing, derivatives, detrending,
       compositions) instead of AOM's ~38 parametric variants. Fewer,
       more diverse operators reduce selection noise.

    Examples
    --------
    >>> from nirs4all.operators.models.sklearn.pop_pls import POPPLSRegressor
    >>> import numpy as np
    >>> rng = np.random.RandomState(42)
    >>> X = rng.randn(100, 200)
    >>> y = X[:, :5].sum(axis=1) + 0.1 * rng.randn(100)
    >>> model = POPPLSRegressor(n_components=10)
    >>> model.fit(X, y)
    POPPLSRegressor(n_components=10)
    >>> preds = model.predict(X)
    >>> # Each component may use a different operator:
    >>> print(model.get_component_operators())

    See Also
    --------
    AOMPLSRegressor : Single-operator selection for all components.
    SIMPLS : Standard SIMPLS regressor.
    """

    _webapp_meta = {
        "category": "pls",
        "tier": "advanced",
        "tags": ["pls", "pop-pls", "per-component", "preprocessing", "auto-select", "regression"],
    }

    _estimator_type = "regressor"

    def __init__(
        self,
        n_components: int = 27,
        operator_bank: list[LinearOperator] | None = None,
        n_orth: int = 0,
        center: bool = True,
        scale: bool = False,
        auto_select: bool = True,
        random_state: int | None = None,
    ):
        self.n_components = n_components
        self.operator_bank = operator_bank
        self.n_orth = n_orth
        self.center = center
        self.scale = scale
        self.auto_select = auto_select
        self.random_state = random_state

    def fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
        X_val: ArrayLike | None = None,
        y_val: ArrayLike | None = None,
    ) -> POPPLSRegressor:
        """Fit the POP-PLS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values.
        X_val : array-like of shape (n_val, n_features), optional
            Validation data for operator and prefix selection. When provided
            (together with y_val), used directly instead of PRESS-based
            internal selection.
        y_val : array-like of shape (n_val,) or (n_val, n_targets), optional
            Validation targets.

        Returns
        -------
        self : POPPLSRegressor
            Fitted estimator.
        """
        X = np.asarray(X, dtype=np.float64)
        y = np.asarray(y, dtype=np.float64)

        self._y_1d = y.ndim == 1
        if self._y_1d:
            y = y.reshape(-1, 1)

        n_samples, n_features = X.shape
        self.n_features_in_ = n_features

        # Limit components by data dimensions
        max_components = min(n_samples - 1, n_features)
        n_comp = min(self.n_components, max_components)

        # Center and optionally scale
        if self.center:
            self.x_mean_ = X.mean(axis=0)
            self.y_mean_ = y.mean(axis=0)
        else:
            self.x_mean_ = np.zeros(n_features, dtype=np.float64)
            self.y_mean_ = np.zeros(y.shape[1], dtype=np.float64)

        if self.scale:
            self.x_std_ = X.std(axis=0, ddof=1)
            self.y_std_ = y.std(axis=0, ddof=1)
            self.x_std_ = np.where(self.x_std_ < 1e-10, 1.0, self.x_std_)
            self.y_std_ = np.where(self.y_std_ < 1e-10, 1.0, self.y_std_)
        else:
            self.x_std_ = np.ones(n_features, dtype=np.float64)
            self.y_std_ = np.ones(y.shape[1], dtype=np.float64)

        X_centered = (X - self.x_mean_) / self.x_std_
        Y_centered = (y - self.y_mean_) / self.y_std_

        # Initialize operator bank
        operators = self.operator_bank if self.operator_bank is not None else pop_pls_operator_bank()
        if not any(isinstance(op, IdentityOperator) for op in operators):
            operators = [IdentityOperator()] + list(operators)
        self.operators_ = list(operators)
        for op in self.operators_:
            op.initialize(n_features)
        self.block_names_ = [op.name for op in self.operators_]

        # Center/scale validation data
        X_val_c = None
        Y_val_c = None
        if X_val is not None and y_val is not None:
            X_v = np.asarray(X_val, dtype=np.float64)
            y_v = np.asarray(y_val, dtype=np.float64)
            if self._y_1d and y_v.ndim == 1:
                y_v = y_v.reshape(-1, 1)
            X_val_c = (X_v - self.x_mean_) / self.x_std_
            Y_val_c = (y_v - self.y_mean_) / self.y_std_

        # Fit
        artifacts = _poppls_fit_numpy(
            X_centered, Y_centered, self.operators_, n_comp, self.n_orth,
            self.auto_select, X_val_c, Y_val_c, self.random_state,
        )

        # Unpack artifacts
        self.n_components_ = artifacts["n_extracted"]
        self.k_selected_ = artifacts["k_selected"]
        self._W = artifacts["W"]
        self._T = artifacts["T"]
        self._P = artifacts["P"]
        self._Q = artifacts["Q"]
        self.gamma_ = artifacts["Gamma"]
        self._B_coefs = artifacts["B_coefs"]
        self._P_orth = artifacts["P_orth"]
        self.component_operators_ = artifacts["component_operators"]

        # Store regression coefficients for selected prefix
        if self.n_components_ > 0:
            B_selected = self._B_coefs[self.k_selected_ - 1]
            self.coef_ = B_selected * self.y_std_[np.newaxis, :] / self.x_std_[:, np.newaxis]
        else:
            self.coef_ = np.zeros((n_features, y.shape[1]), dtype=np.float64)

        return self

    def predict(
        self,
        X: ArrayLike,
        n_components: int | None = None,
    ) -> NDArray[np.floating]:
        """Predict using the POP-PLS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples to predict.
        n_components : int, optional
            Number of components to use. If None, uses k_selected_.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,) or (n_samples, n_targets)
            Predicted values.
        """
        check_is_fitted(self, ["x_mean_", "x_std_", "y_mean_", "y_std_", "_B_coefs"])

        X = np.asarray(X, dtype=np.float64)
        X_centered = (X - self.x_mean_) / self.x_std_

        # Apply OPLS filter
        if self._P_orth is not None:
            for j in range(self._P_orth.shape[1]):
                p_o = self._P_orth[:, j]
                t_o = X_centered @ p_o
                X_centered = X_centered - np.outer(t_o, p_o)

        if n_components is None:
            n_components = self.k_selected_
        n_components = min(n_components, self.n_components_)

        y_pred: NDArray[np.floating]
        if n_components == 0:
            y_pred = np.full((X.shape[0], len(self.y_mean_)), self.y_mean_, dtype=np.float64)
        else:
            B_k = self._B_coefs[n_components - 1]
            y_pred_std = X_centered @ B_k
            y_pred = y_pred_std * self.y_std_ + self.y_mean_

        if self._y_1d:
            y_pred = y_pred.ravel()
        return y_pred

    def transform(self, X: ArrayLike) -> NDArray[np.floating]:
        """Transform X to score space.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples to transform.

        Returns
        -------
        T : ndarray of shape (n_samples, k_selected_)
            X scores.
        """
        check_is_fitted(self, ["x_mean_", "x_std_", "_W"])

        X = np.asarray(X, dtype=np.float64)
        X_centered = (X - self.x_mean_) / self.x_std_

        if self._P_orth is not None:
            for j in range(self._P_orth.shape[1]):
                p_o = self._P_orth[:, j]
                t_o = X_centered @ p_o
                X_centered = X_centered - np.outer(t_o, p_o)

        return np.asarray(X_centered @ self._W[:, :self.k_selected_])

    def get_block_weights(self) -> NDArray[np.floating]:
        """Get per-component operator selection matrix.

        Returns
        -------
        gamma : ndarray of shape (n_components_, n_blocks)
            Selection matrix. Each row has exactly one 1.0 entry indicating
            which operator was selected for that component.
        """
        check_is_fitted(self, ["gamma_"])
        return np.array(self.gamma_, copy=True)

    def get_preprocessing_report(self) -> list[dict]:
        """Get a human-readable report of per-component operator selections.

        Returns
        -------
        report : list of dict
            One entry per component with fields: 'component', 'blocks'
            (list of {name, weight} dicts for non-zero blocks).
        """
        check_is_fitted(self, ["gamma_", "block_names_"])
        report: list[dict] = []
        for k in range(self.n_components_):
            blocks: list[dict[str, str | float]] = []
            for b, name in enumerate(self.block_names_):
                if self.gamma_[k, b] > 1e-6:
                    blocks.append({"name": name, "weight": float(self.gamma_[k, b])})
            blocks.sort(key=lambda x: float(x["weight"]), reverse=True)
            report.append({"component": k + 1, "blocks": blocks})
        return report

    def get_component_operators(self) -> list[str]:
        """Get the operator name selected for each component.

        Returns
        -------
        operators : list of str
            Operator name for each extracted component (length n_components_).
        """
        check_is_fitted(self, ["component_operators_"])
        return list(self.component_operators_)

    def get_params(self, deep: bool = True) -> dict:
        """Get parameters for this estimator."""
        return {
            "n_components": self.n_components,
            "operator_bank": self.operator_bank,
            "n_orth": self.n_orth,
            "center": self.center,
            "scale": self.scale,
            "auto_select": self.auto_select,
            "random_state": self.random_state,
        }

    def set_params(self, **params) -> POPPLSRegressor:
        """Set the parameters of this estimator."""
        for key, value in params.items():
            setattr(self, key, value)
        return self

    def __repr__(self) -> str:
        return (
            f"POPPLSRegressor(n_components={self.n_components}, "
            f"n_orth={self.n_orth}, auto_select={self.auto_select})"
        )



import operator

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class Rotate_Translate(TransformerMixin, BaseEstimator):
    """
    Class for rotating and translating data augmentation.

    Vectorized implementation that processes all samples in batch.

    Parameters
    ----------
    p_range : int, optional
        Range for generating random slope values. Default is 2.
    y_factor : int, optional
        Scaling factor for the initial value. Default is 3.
    random_state : int or None, optional
        Random seed for reproducibility. Default is None.
    """

    _webapp_meta = {
        "category": "random",
        "tier": "standard",
        "tags": ["random", "rotate", "translate", "augmentation"],
    }

    def __init__(self, p_range=2, y_factor=3, random_state=None):
        self.p_range = p_range
        self.y_factor = y_factor
        self.random_state = random_state

    def fit(self, X, y=None, **kwargs):
        self._rng = np.random.default_rng(self.random_state)
        return self

    def transform(self, X, **kwargs):
        """
        Transform the data by rotating and translating the signal.

        Vectorized implementation using NumPy broadcasting.

        Parameters
        ----------
        X : ndarray
            Input data to be transformed, shape (n_samples, n_features).

        Returns
        -------
        ndarray
            Transformed data.
        """
        rng = getattr(self, '_rng', np.random.default_rng(self.random_state))
        n_samples, n_features = X.shape

        # Pre-compute x_range once for all samples
        x_range = np.linspace(0, 1, n_features)  # (n_features,)

        # Generate all random parameters at once for all samples
        p1 = rng.uniform(-self.p_range / 5, self.p_range / 5, n_samples)  # (n_samples,)
        p2 = rng.uniform(-self.p_range / 5, self.p_range / 5, n_samples)  # (n_samples,)
        xI = rng.uniform(0, 1, n_samples)  # (n_samples,)

        # Compute yI for each sample based on max values
        max_vals = np.max(X, axis=1)  # (n_samples,)
        yI_upper = np.maximum(0, max_vals / self.y_factor)
        yI = rng.uniform(0, 1, n_samples) * yI_upper  # (n_samples,)

        # Vectorized computation using broadcasting
        # x_range: (n_features,), xI: (n_samples,) -> broadcast to (n_samples, n_features)
        x_expanded = x_range[np.newaxis, :]  # (1, n_features)
        xI_expanded = xI[:, np.newaxis]  # (n_samples, 1)

        # Compute mask for each sample
        mask = x_expanded <= xI_expanded  # (n_samples, n_features)

        # Compute slopes for both branches using broadcasting
        p1_expanded = p1[:, np.newaxis]  # (n_samples, 1)
        p2_expanded = p2[:, np.newaxis]  # (n_samples, 1)
        yI_expanded = yI[:, np.newaxis]  # (n_samples, 1)

        # Vectorized angle computation for all samples at once
        distor = np.where(
            mask,
            p1_expanded * (x_expanded - xI_expanded) + yI_expanded,
            p2_expanded * (x_expanded - xI_expanded) + yI_expanded
        )  # (n_samples, n_features)

        # Multiply by per-sample std
        stds = np.std(X, axis=1, keepdims=True)  # (n_samples, 1)
        increment = distor * stds  # (n_samples, n_features)

        return X + increment

class Random_X_Operation(TransformerMixin, BaseEstimator):
    """
    Class for applying random operation on data augmentation.

    Parameters
    ----------
    operator_func : function, optional
        Operator function to be applied. Default is operator.mul.
    operator_range : tuple, optional
        Range for generating random values for the operator. Default is (0.97, 1.03).
    random_state : int or None, optional
        Random seed for reproducibility. Default is None.
    """

    _webapp_meta = {
        "category": "random",
        "tier": "standard",
        "tags": ["random", "operation", "multiplicative", "augmentation"],
    }

    def __init__(self, operator_func=operator.mul, operator_range=(0.97, 1.03), random_state=None):
        self.operator_func = operator_func
        self.operator_range = operator_range
        self.random_state = random_state

    def fit(self, X, y=None, **kwargs):
        self._rng = np.random.default_rng(self.random_state)
        return self

    def transform(self, X, **kwargs):
        """
        Transform the data by applying random operation.

        Parameters
        ----------
        X : ndarray
            Input data to be transformed.

        Returns
        -------
        ndarray
            Transformed data.
        """
        rng = getattr(self, '_rng', np.random.default_rng(self.random_state))
        min_val = self.operator_range[0]
        interval = self.operator_range[1] - self.operator_range[0]

        increment = rng.random(X.shape) * interval + min_val

        new_X = self.operator_func(X, increment)
        # Clip the augmented data within the float32 range
        new_X = np.clip(new_X, -np.finfo(np.float32).max, np.finfo(np.float32).max)

        return new_X



"""
Validators for dataset configuration.

This module provides validation logic for dataset configurations,
checking for consistency, required fields, file existence, and other rules.
"""

from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np


@dataclass
class ValidationError:
    """Represents a validation error.

    Attributes:
        code: Error code for programmatic handling.
        message: Human-readable error message.
        field: The configuration field that caused the error.
        value: The value that caused the error.
        suggestion: Optional suggestion for fixing the error.
    """

    code: str
    message: str
    field: str | None = None
    value: Any = None
    suggestion: str | None = None

    def __str__(self) -> str:
        parts = [self.message]
        if self.field:
            parts.insert(0, f"[{self.field}]")
        if self.suggestion:
            parts.append(f"Suggestion: {self.suggestion}")
        return " ".join(parts)

@dataclass
class ValidationWarning:
    """Represents a validation warning (non-fatal issue).

    Attributes:
        code: Warning code for programmatic handling.
        message: Human-readable warning message.
        field: The configuration field that caused the warning.
    """

    code: str
    message: str
    field: str | None = None

    def __str__(self) -> str:
        if self.field:
            return f"[{self.field}] {self.message}"
        return self.message

@dataclass
class ValidationResult:
    """Result of configuration validation.

    Attributes:
        is_valid: Whether the configuration is valid (no errors).
        errors: List of validation errors.
        warnings: List of validation warnings.
        normalized_config: The validated and normalized configuration.
    """

    is_valid: bool
    errors: list[ValidationError] = field(default_factory=list)
    warnings: list[ValidationWarning] = field(default_factory=list)
    normalized_config: dict[str, Any] | None = None

    def __str__(self) -> str:
        if self.is_valid:
            msg = "Configuration is valid"
            if self.warnings:
                msg += f" with {len(self.warnings)} warning(s)"
            return msg
        return f"Configuration is invalid: {len(self.errors)} error(s)"

    def raise_if_invalid(self) -> None:
        """Raise ValueError if configuration is invalid."""
        if not self.is_valid:
            error_messages = [str(e) for e in self.errors]
            raise ValueError(
                "Invalid configuration:\n" +
                "\n".join(f"  - {msg}" for msg in error_messages)
            )

class ConfigValidator:
    """Validator for dataset configurations.

    Provides validation rules and methods for checking dataset configurations.
    Supports both legacy and new format configurations.

    Example:
        ```python
        validator = ConfigValidator()
        result = validator.validate(config_dict)
        if not result.is_valid:
            for error in result.errors:
                print(f"Error: {error}")
        ```
    """

    def __init__(
        self,
        check_file_existence: bool = False,
        custom_validators: list[Callable] | None = None
    ):
        """Initialize the validator.

        Args:
            check_file_existence: Whether to check if referenced files exist.
                Default is False to allow validation before files are available.
            custom_validators: Optional list of custom validation functions.
                Each function should accept (config, errors, warnings) and
                add any issues to the lists.
        """
        self.check_file_existence = check_file_existence
        self.custom_validators = custom_validators or []

    def validate(self, config: dict[str, Any]) -> ValidationResult:
        """Validate a configuration dictionary.

        Args:
            config: Configuration dictionary to validate.

        Returns:
            ValidationResult with errors, warnings, and normalized config.
        """
        errors: list[ValidationError] = []
        warnings: list[ValidationWarning] = []

        # Create a copy for normalization
        normalized = dict(config)

        # Run validation rules
        self._validate_data_sources(normalized, errors, warnings)
        self._validate_task_type(normalized, errors, warnings)
        self._validate_loading_params(normalized, errors, warnings)
        self._validate_aggregation(normalized, errors, warnings)

        if self.check_file_existence:
            self._validate_file_existence(normalized, errors, warnings)

        # Run custom validators
        for validator in self.custom_validators:
            validator(normalized, errors, warnings)

        return ValidationResult(
            is_valid=len(errors) == 0,
            errors=errors,
            warnings=warnings,
            normalized_config=normalized if len(errors) == 0 else None
        )

    def _validate_data_sources(
        self,
        config: dict[str, Any],
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate that data sources are properly specified."""
        has_train_x = config.get("train_x") is not None
        has_test_x = config.get("test_x") is not None
        has_files = config.get("files") is not None
        has_sources = config.get("sources") is not None

        # Check for any data source
        if not has_train_x and not has_test_x and not has_files and not has_sources:
            errors.append(ValidationError(
                code="NO_DATA_SOURCE",
                message="No data source specified. "
                        "Provide train_x, test_x, files, or sources.",
                suggestion="Add train_x with path to training features CSV."
            ))

        # Check for mixed formats (warning only)
        if (has_train_x or has_test_x) and (has_files or has_sources):
            warnings.append(ValidationWarning(
                code="MIXED_FORMAT",
                message="Both train_x/test_x and files/sources detected. "
                        "Use one format consistently."
            ))

        # Validate multi-source consistency
        train_x = config.get("train_x")
        if isinstance(train_x, list) and not all(isinstance(p, (str, Path)) for p in train_x) and not all(isinstance(p, np.ndarray) for p in train_x):
            errors.append(ValidationError(
                    code="MIXED_SOURCE_TYPES",
                    message="Multi-source train_x contains mixed types.",
                    field="train_x",
                    suggestion="Use either all file paths or all numpy arrays."
                ))

    def _validate_task_type(
        self,
        config: dict[str, Any],
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate task_type configuration."""
        task_type = config.get("task_type")
        if task_type is not None:
            valid_types = ["auto", "regression", "binary_classification", "multiclass_classification"]
            if isinstance(task_type, str) and task_type.lower() not in valid_types:
                errors.append(ValidationError(
                    code="INVALID_TASK_TYPE",
                    message=f"Invalid task_type: '{task_type}'",
                    field="task_type",
                    value=task_type,
                    suggestion=f"Valid values: {valid_types}"
                ))

    def _validate_loading_params(
        self,
        config: dict[str, Any],
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate loading parameters."""
        # Check global_params
        global_params = config.get("global_params")
        if global_params is not None:
            self._validate_params_dict(global_params, "global_params", errors, warnings)

        # Check partition-level params
        for partition in ["train", "test"]:
            params_key = f"{partition}_params"
            params = config.get(params_key)
            if params is not None:
                self._validate_params_dict(params, params_key, errors, warnings)

            # Check file-level params
            for data_type in ["x", "y", "group"]:
                params_key = f"{partition}_{data_type}_params"
                params = config.get(params_key)
                if params is not None:
                    self._validate_params_dict(params, params_key, errors, warnings)

    def _validate_params_dict(
        self,
        params: Any,
        field_name: str,
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate a params dictionary."""
        if not isinstance(params, dict):
            errors.append(ValidationError(
                code="INVALID_PARAMS_TYPE",
                message=f"Expected dict for {field_name}, got {type(params).__name__}",
                field=field_name,
                value=params
            ))
            return

        # Validate header_unit
        header_unit = params.get("header_unit")
        if header_unit is not None:
            valid_units = ["cm-1", "nm", "none", "text", "index"]
            if isinstance(header_unit, str) and header_unit.lower() not in valid_units:
                errors.append(ValidationError(
                    code="INVALID_HEADER_UNIT",
                    message=f"Invalid header_unit: '{header_unit}'",
                    field=f"{field_name}.header_unit",
                    value=header_unit,
                    suggestion=f"Valid values: {valid_units}"
                ))

        # Validate signal_type
        signal_type = params.get("signal_type")
        if signal_type is not None:
            valid_types = [
                "auto", "absorbance", "reflectance", "reflectance%",
                "transmittance", "transmittance%", "log(1/R)", "kubelka-munk"
            ]
            if isinstance(signal_type, str) and signal_type.lower() not in valid_types:
                warnings.append(ValidationWarning(
                    code="UNKNOWN_SIGNAL_TYPE",
                    message=f"Unknown signal_type: '{signal_type}'. May be auto-detected.",
                    field=f"{field_name}.signal_type"
                ))

        # Validate na_policy
        na_policy = params.get("na_policy")
        if na_policy is not None:
            valid_policies = ["auto", "abort", "remove_sample", "remove_feature", "replace", "ignore"]
            if isinstance(na_policy, str) and na_policy.lower() not in valid_policies:
                errors.append(ValidationError(
                    code="INVALID_NA_POLICY",
                    message=f"Invalid na_policy: '{na_policy}'",
                    field=f"{field_name}.na_policy",
                    value=na_policy,
                    suggestion=f"Valid values: {valid_policies}"
                ))

    def _validate_aggregation(
        self,
        config: dict[str, Any],
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate aggregation settings."""
        aggregate = config.get("aggregate")
        aggregate_method = config.get("aggregate_method")

        # aggregate_method without aggregate
        if aggregate_method is not None and aggregate is None:
            warnings.append(ValidationWarning(
                code="UNUSED_AGGREGATE_METHOD",
                message="aggregate_method specified without aggregate. It will be ignored.",
                field="aggregate_method"
            ))

        # Validate aggregate_method value
        if aggregate_method is not None:
            valid_methods = ["mean", "median", "vote"]
            if isinstance(aggregate_method, str) and aggregate_method.lower() not in valid_methods:
                errors.append(ValidationError(
                    code="INVALID_AGGREGATE_METHOD",
                    message=f"Invalid aggregate_method: '{aggregate_method}'",
                    field="aggregate_method",
                    value=aggregate_method,
                    suggestion=f"Valid values: {valid_methods}"
                ))

    def _validate_file_existence(
        self,
        config: dict[str, Any],
        errors: list[ValidationError],
        warnings: list[ValidationWarning]
    ) -> None:
        """Validate that referenced files exist."""
        file_fields = [
            "train_x", "train_y", "train_group",
            "test_x", "test_y", "test_group"
        ]

        for field_name in file_fields:
            value = config.get(field_name)
            if value is None:
                continue

            # Skip numpy arrays
            if isinstance(value, np.ndarray):
                continue

            # Handle lists (multi-source)
            if isinstance(value, list):
                for i, path in enumerate(value):
                    if isinstance(path, (str, Path)) and not Path(path).exists():
                        warnings.append(ValidationWarning(
                            code="FILE_NOT_FOUND",
                            message=f"File not found: {path}",
                            field=f"{field_name}[{i}]"
                        ))
            elif isinstance(value, (str, Path)) and not Path(value).exists():
                warnings.append(ValidationWarning(
                    code="FILE_NOT_FOUND",
                    message=f"File not found: {value}",
                    field=field_name
                ))

def validate_config(
    config: dict[str, Any],
    check_file_existence: bool = False
) -> ValidationResult:
    """Convenience function to validate a configuration.

    Args:
        config: Configuration dictionary to validate.
        check_file_existence: Whether to check if referenced files exist.

    Returns:
        ValidationResult with errors, warnings, and normalized config.
    """
    validator = ConfigValidator(check_file_existence=check_file_existence)
    return validator.validate(config)



"""Multiblock PLS (MB-PLS) regressor for nirs4all.

MB-PLS fuses multiple X blocks (e.g., different preprocessing variants,
multiple sensors) into a single predictive model. Each block contributes
to the latent variables according to its relevance to Y.
"""

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin


def _check_jax_available():
    """Check if JAX is available for GPU acceleration."""
    try:
        import jax
        return True
    except ImportError:
        return False

# =============================================================================
# NumPy Backend Implementation
# =============================================================================

def _mbpls_fit_numpy(X, y, n_components, standardize=True):
    """Fit single-block MBPLS using pure NumPy (NIPALS algorithm).

    Parameters
    ----------
    X : ndarray of shape (n_samples, n_features)
        Training data.
    y : ndarray of shape (n_samples,) or (n_samples, n_targets)
        Target values.
    n_components : int
        Number of components to extract.
    standardize : bool, default=True
        Whether to standardize X and y.

    Returns
    -------
    B : ndarray
        Regression coefficients.
    W : ndarray
        Weight matrix.
    P : ndarray
        X loading matrix.
    Q : ndarray
        Y loading matrix.
    T : ndarray
        Score matrix.
    X_mean, X_std, y_mean, y_std : ndarray
        Preprocessing parameters.
    """
    X = np.asarray(X, dtype=np.float64)
    y = np.asarray(y, dtype=np.float64)

    n_samples, n_features = X.shape

    # Center and scale
    X_mean = np.mean(X, axis=0, keepdims=True)
    if standardize:
        X_std = np.std(X, axis=0, keepdims=True, ddof=1)
        X_std = np.where(X_std < 1e-10, 1.0, X_std)
    else:
        X_std = np.ones((1, n_features), dtype=np.float64)
    X_centered = (X - X_mean) / X_std

    y = y.reshape(-1, 1) if y.ndim == 1 else y
    n_targets = y.shape[1]
    y_mean = np.mean(y, axis=0, keepdims=True)
    if standardize:
        y_std = np.std(y, axis=0, keepdims=True, ddof=1)
        y_std = np.where(y_std < 1e-10, 1.0, y_std)
    else:
        y_std = np.ones((1, n_targets), dtype=np.float64)
    y_centered = (y - y_mean) / y_std

    # Initialize matrices
    W = np.zeros((n_features, n_components), dtype=np.float64)
    P = np.zeros((n_features, n_components), dtype=np.float64)
    Q = np.zeros((n_targets, n_components), dtype=np.float64)
    T = np.zeros((n_samples, n_components), dtype=np.float64)

    X_res = X_centered.copy()
    y_res = y_centered.copy()

    for i in range(n_components):
        # Weight vector
        w = X_res.T @ y_res
        if n_targets == 1:
            w = w.ravel()
        else:
            # For multi-target, use first column
            w = w[:, 0]
        w_norm = np.linalg.norm(w)
        if w_norm > 1e-10:
            w = w / w_norm

        # Scores
        t = X_res @ w
        t_norm = t.T @ t + 1e-10

        # Loadings
        p = X_res.T @ t / t_norm
        q = y_res.T @ t / t_norm

        # Store
        W[:, i] = w
        P[:, i] = p
        Q[:, i] = q.ravel()
        T[:, i] = t

        # Deflate
        X_res = X_res - np.outer(t, p)
        y_res = y_res - np.outer(t, q)

    # Compute final regression coefficients: B = W @ inv(P.T @ W) @ Q.T
    PtW = P.T @ W
    PtW_reg = PtW + 1e-10 * np.eye(n_components)
    PtW_inv = np.linalg.pinv(PtW_reg)
    B_final = W @ PtW_inv @ Q.T

    return B_final, W, P, Q, T, X_mean, X_std, y_mean, y_std

def _mbpls_fit_multiblock_numpy(X_blocks, y, n_components, standardize=True):
    """Fit multiblock MBPLS using pure NumPy.

    Parameters
    ----------
    X_blocks : list of ndarray
        List of X blocks, each of shape (n_samples, n_features_block).
    y : ndarray of shape (n_samples,) or (n_samples, n_targets)
        Target values.
    n_components : int
        Number of components to extract.
    standardize : bool, default=True
        Whether to standardize blocks.

    Returns
    -------
    B : ndarray
        Regression coefficients for concatenated features.
    W : ndarray
        Weight matrix.
    P : ndarray
        X loading matrix.
    Q : ndarray
        Y loading matrix.
    T : ndarray
        Super score matrix.
    block_weights : list of ndarray
        Block-level weight matrices.
    X_means, X_stds : list of ndarray
        Preprocessing parameters for each block.
    y_mean, y_std : ndarray
        Preprocessing parameters for y.
    """
    n_blocks = len(X_blocks)
    n_samples = X_blocks[0].shape[0]

    y = np.asarray(y, dtype=np.float64)
    y = y.reshape(-1, 1) if y.ndim == 1 else y
    n_targets = y.shape[1]

    # Preprocess y
    y_mean = np.mean(y, axis=0, keepdims=True)
    if standardize:
        y_std = np.std(y, axis=0, keepdims=True, ddof=1)
        y_std = np.where(y_std < 1e-10, 1.0, y_std)
    else:
        y_std = np.ones((1, n_targets), dtype=np.float64)
    y_centered = (y - y_mean) / y_std

    # Preprocess each block
    X_centered_blocks = []
    X_means = []
    X_stds = []
    block_sizes = []

    for X_block in X_blocks:
        X_block = np.asarray(X_block, dtype=np.float64)
        n_features_block = X_block.shape[1]
        block_sizes.append(n_features_block)

        X_mean = np.mean(X_block, axis=0, keepdims=True)
        if standardize:
            X_std = np.std(X_block, axis=0, keepdims=True, ddof=1)
            X_std = np.where(X_std < 1e-10, 1.0, X_std)
        else:
            X_std = np.ones((1, n_features_block), dtype=np.float64)

        X_means.append(X_mean)
        X_stds.append(X_std)
        X_centered_blocks.append((X_block - X_mean) / X_std)

    total_features = sum(block_sizes)

    # Initialize matrices
    W = np.zeros((total_features, n_components), dtype=np.float64)
    P = np.zeros((total_features, n_components), dtype=np.float64)
    Q = np.zeros((n_targets, n_components), dtype=np.float64)
    T = np.zeros((n_samples, n_components), dtype=np.float64)
    block_weights = [np.zeros((bs, n_components), dtype=np.float64) for bs in block_sizes]

    X_res_blocks = [X.copy() for X in X_centered_blocks]
    y_res = y_centered.copy()

    for comp in range(n_components):
        # Compute block scores and weights
        block_scores = []
        for b, X_res in enumerate(X_res_blocks):
            # Block weight
            w_b = X_res.T @ y_res
            w_b = w_b.ravel() if n_targets == 1 else w_b[:, 0]
            w_norm = np.linalg.norm(w_b)
            if w_norm > 1e-10:
                w_b = w_b / w_norm

            # Block score
            t_b = X_res @ w_b
            block_scores.append(t_b)
            block_weights[b][:, comp] = w_b

        # Super score (average of block scores)
        t = np.mean(block_scores, axis=0)
        t_norm = np.linalg.norm(t)
        if t_norm > 1e-10:
            t = t / t_norm
        t_dot = t.T @ t + 1e-10

        T[:, comp] = t

        # Compute loadings for each block and concatenate
        p_concat = []
        for _b, X_res in enumerate(X_res_blocks):
            p_b = X_res.T @ t / t_dot
            p_concat.append(p_b)

        p_full = np.concatenate(p_concat)
        P[:, comp] = p_full

        # Y loading
        q = y_res.T @ t / t_dot
        Q[:, comp] = q.ravel()

        # Concatenated weight
        w_full = np.concatenate([block_weights[b][:, comp] for b in range(n_blocks)])
        W[:, comp] = w_full

        # Deflate each block
        for b, X_res in enumerate(X_res_blocks):
            p_b = p_concat[b]
            X_res_blocks[b] = X_res - np.outer(t, p_b)

        y_res = y_res - np.outer(t, q)

    # Compute final regression coefficients
    PtW = P.T @ W
    PtW_reg = PtW + 1e-10 * np.eye(n_components)
    PtW_inv = np.linalg.pinv(PtW_reg)
    B_final = W @ PtW_inv @ Q.T

    return B_final, W, P, Q, T, block_weights, X_means, X_stds, y_mean, y_std

def _mbpls_predict_numpy(X, B, X_mean, X_std, y_mean, y_std):
    """Predict using MBPLS coefficients (NumPy)."""
    X = np.asarray(X, dtype=np.float64)
    X_centered = (X - X_mean) / X_std
    y_pred_centered = X_centered @ B
    return y_pred_centered * y_std + y_mean

def _mbpls_predict_multiblock_numpy(X_blocks, B, X_means, X_stds, y_mean, y_std):
    """Predict using multiblock MBPLS coefficients (NumPy)."""
    # Preprocess and concatenate blocks
    X_centered_list = []
    for b, X_block in enumerate(X_blocks):
        X_block = np.asarray(X_block, dtype=np.float64)
        X_centered = (X_block - X_means[b]) / X_stds[b]
        X_centered_list.append(X_centered)

    X_concat = np.hstack(X_centered_list)
    y_pred_centered = X_concat @ B
    return y_pred_centered * y_std + y_mean

def _mbpls_transform_numpy(X, W, X_mean, X_std):
    """Transform X to latent space (NumPy)."""
    X = np.asarray(X, dtype=np.float64)
    X_centered = (X - X_mean) / X_std
    return X_centered @ W

def _mbpls_transform_multiblock_numpy(X_blocks, W, X_means, X_stds, block_sizes):
    """Transform multiblock X to latent space (NumPy)."""
    X_centered_list = []
    for b, X_block in enumerate(X_blocks):
        X_block = np.asarray(X_block, dtype=np.float64)
        X_centered = (X_block - X_means[b]) / X_stds[b]
        X_centered_list.append(X_centered)

    X_concat = np.hstack(X_centered_list)
    return X_concat @ W

# =============================================================================
# JAX Backend Implementations
# =============================================================================

def _get_jax_mbpls_functions():
    """Get JAX-accelerated MBPLS functions (single block)."""
    from functools import partial

    import jax
    import jax.numpy as jnp
    from jax import lax

    jax.config.update("jax_enable_x64", True)

    @partial(jax.jit, static_argnums=(2,))
    def mbpls_fit_jax(X, y, n_components):
        """Fit single-block MBPLS using JAX (equivalent to NIPALS PLS).

        Returns
        -------
        B : Regression coefficients for each component
        W : Weight matrix
        P : Loading matrix
        Q : Y loading matrix
        T : Score matrix
        X_mean, X_std, y_mean, y_std : Preprocessing parameters
        """
        # Ensure float64 for numerical precision
        X = jnp.asarray(X, dtype=jnp.float64)
        y = jnp.asarray(y, dtype=jnp.float64)

        n_samples, n_features = X.shape

        # Center and scale
        X_mean = jnp.mean(X, axis=0, keepdims=True)
        X_std = jnp.std(X, axis=0, keepdims=True, ddof=1)
        X_std = jnp.where(X_std < 1e-10, 1.0, X_std)
        X_centered = (X - X_mean) / X_std

        y = y.reshape(-1, 1) if y.ndim == 1 else y
        n_targets = y.shape[1]
        y_mean = jnp.mean(y, axis=0, keepdims=True)
        y_std = jnp.std(y, axis=0, keepdims=True, ddof=1)
        y_std = jnp.where(y_std < 1e-10, 1.0, y_std)
        y_centered = (y - y_mean) / y_std

        # Initialize matrices with explicit dtype
        W = jnp.zeros((n_features, n_components), dtype=jnp.float64)
        P = jnp.zeros((n_features, n_components), dtype=jnp.float64)
        Q = jnp.zeros((n_targets, n_components), dtype=jnp.float64)
        T = jnp.zeros((n_samples, n_components), dtype=jnp.float64)

        def component_step(i, carry):
            X_res, y_res, W, P, Q, T = carry

            # Weight vector
            w = X_res.T @ y_res
            if n_targets == 1:
                w = w.ravel()
            else:
                # For multi-target, use first singular vector
                w = w[:, 0]
            w = w / (jnp.linalg.norm(w) + 1e-10)

            # Scores
            t = X_res @ w
            t_norm = t.T @ t + 1e-10

            # Loadings
            p = X_res.T @ t / t_norm
            q = y_res.T @ t / t_norm

            # Store
            W = W.at[:, i].set(w)
            P = P.at[:, i].set(p)
            Q = Q.at[:, i].set(q.ravel())
            T = T.at[:, i].set(t)

            # Deflate
            X_res = X_res - jnp.outer(t, p)
            y_res = y_res - jnp.outer(t, q)

            return X_res, y_res, W, P, Q, T

        _, _, W, P, Q, T = lax.fori_loop(
            0, n_components, component_step,
            (X_centered, y_centered, W, P, Q, T)
        )

        # Compute final regression coefficients: B = W @ inv(P.T @ W) @ Q.T
        PtW = P.T @ W
        PtW_reg = PtW + 1e-10 * jnp.eye(n_components)
        PtW_inv = jnp.linalg.pinv(PtW_reg)
        B_final = W @ PtW_inv @ Q.T

        return B_final, W, P, Q, T, X_mean, X_std, y_mean, y_std

    @jax.jit
    def mbpls_predict_jax(X, B, X_mean, X_std, y_mean, y_std):
        """Predict using MBPLS coefficients."""
        X = jnp.asarray(X, dtype=jnp.float64)
        X_centered = (X - X_mean) / X_std
        y_pred_centered = X_centered @ B
        return y_pred_centered * y_std + y_mean

    return mbpls_fit_jax, mbpls_predict_jax

# Cache JAX functions
_JAX_MBPLS_FUNCS = None

def _get_cached_jax_mbpls():
    """Get cached JAX MBPLS functions."""
    global _JAX_MBPLS_FUNCS
    if _JAX_MBPLS_FUNCS is None:
        _JAX_MBPLS_FUNCS = _get_jax_mbpls_functions()
    return _JAX_MBPLS_FUNCS

class MBPLS(BaseEstimator, RegressorMixin):
    """Multiblock PLS (MB-PLS) regressor.

    MB-PLS fuses multiple X blocks (e.g., different preprocessing variants,
    multiple sensors) into a single predictive model. Each block contributes
    to the latent variables according to its relevance to Y.

    Parameters
    ----------
    n_components : int, default=5
        Number of latent variables to extract.
    method : str, default='NIPALS'
        Decomposition method. Currently only 'NIPALS' is supported.
    standardize : bool, default=True
        Whether to standardize blocks before fitting.
    max_tol : float, default=1e-14
        Convergence tolerance for NIPALS.
    backend : str, default='numpy'
        Backend to use for computation. Options are:
        - 'numpy': Use NumPy backend (CPU only).
        - 'jax': Use JAX backend (supports GPU/TPU acceleration).
          Note: JAX backend only supports single-block mode.
        JAX backend requires JAX: ``pip install jax``
        For GPU support: ``pip install jax[cuda12]``

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during fit.
    n_components_ : int
        Actual number of components used.
    coef_ : ndarray of shape (n_features, n_targets)
        Regression coefficients.

    Examples
    --------
    >>> from nirs4all.operators.models.sklearn.pls import MBPLS
    >>> import numpy as np
    >>> X = np.random.randn(100, 50)
    >>> y = np.random.randn(100)
    >>> model = MBPLS(n_components=5)
    >>> model.fit(X, y)
    MBPLS(n_components=5)
    >>> predictions = model.predict(X)
    >>> # Multiblock usage
    >>> X1 = np.random.randn(100, 30)
    >>> X2 = np.random.randn(100, 20)
    >>> model.fit([X1, X2], y)
    >>> # JAX backend for GPU acceleration
    >>> model_jax = MBPLS(n_components=5, backend='jax')

    Notes
    -----
    For JAX with GPU support: ``pip install jax[cuda12]``

    See Also
    --------
    sklearn.cross_decomposition.PLSRegression : Standard single-block PLS.

    References
    ----------
    .. [1] Westerhuis, J. A., et al. (1998). Analysis of multiblock and
           hierarchical PCA and PLS models. Journal of Chemometrics, 12(5),
           301-321.
    """

    _webapp_meta = {
        "category": "pls",
        "tier": "advanced",
        "tags": ["pls", "multiblock", "fusion", "regression"],
    }

    # Explicitly declare estimator type for sklearn compatibility (e.g., StackingRegressor)
    _estimator_type = "regressor"

    def __init__(
        self,
        n_components: int = 5,
        method: str = 'NIPALS',
        standardize: bool = True,
        max_tol: float = 1e-14,
        backend: str = 'numpy',
    ):
        """Initialize MBPLS regressor.

        Parameters
        ----------
        n_components : int, default=5
            Number of latent variables to extract.
        method : str, default='NIPALS'
            Decomposition method (NumPy backend only).
        standardize : bool, default=True
            Whether to standardize blocks before fitting.
        max_tol : float, default=1e-14
            Convergence tolerance for NIPALS.
        backend : str, default='numpy'
            Backend to use ('numpy' or 'jax').
        """
        self.n_components = n_components
        self.method = method
        self.standardize = standardize
        self.max_tol = max_tol
        self.backend = backend

    def fit(self, X, y):
        """Fit the MB-PLS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or list of arrays
            Training data. Can be a single matrix or a list of X blocks
            for true multiblock analysis (NumPy backend only).
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : MBPLS
            Fitted estimator.

        Raises
        ------
        ImportError
            If mbpls package is not installed (NumPy backend),
            or JAX is not available (JAX backend).
        ValueError
            If backend is not 'numpy' or 'jax', or if multiblock
            input is used with JAX backend.
        """
        # Validate backend
        if self.backend not in ('numpy', 'jax'):
            raise ValueError(
                f"backend must be 'numpy' or 'jax', got '{self.backend}'"
            )

        # Handle single array or list of blocks
        if isinstance(X, list):
            if self.backend == 'jax':
                raise ValueError(
                    "JAX backend only supports single-block mode. "
                    "Use backend='numpy' for multiblock analysis."
                )
            X_blocks = [np.asarray(x) for x in X]
            self.n_features_in_ = sum(x.shape[1] for x in X_blocks)
            self._is_multiblock = True
            self._block_sizes = [x.shape[1] for x in X_blocks]
        else:
            X = np.asarray(X)
            X_blocks = [X]
            self.n_features_in_ = X.shape[1]
            self._is_multiblock = False
            self._block_sizes = [X.shape[1]]

        y = np.asarray(y)
        if y.ndim == 1:
            y = y.reshape(-1, 1)

        # Limit components
        n_samples = X_blocks[0].shape[0]
        max_components = min(n_samples - 1, self.n_features_in_)
        self.n_components_ = min(self.n_components, max_components)

        if self.backend == 'jax':
            if not _check_jax_available():
                raise ImportError(
                    "JAX is required for MBPLS with backend='jax'. "
                    "Install it with: pip install jax\n"
                    "For GPU support: pip install jax[cuda12]"
                )

            import jax.numpy as jnp

            # Get JAX functions
            mbpls_fit_jax, _ = _get_cached_jax_mbpls()

            # Fit using JAX
            X_jax = jnp.asarray(X_blocks[0])
            y_jax = jnp.asarray(y)

            result = mbpls_fit_jax(X_jax, y_jax, self.n_components_)
            (self._B, self._W, self._P, self._Q, self._T,
             self._X_mean, self._X_std,
             self._y_mean, self._y_std) = result

            # Store coefficients
            self.coef_ = np.asarray(self._B)
        else:
            # NumPy backend - native implementation
            if self._is_multiblock:
                result = _mbpls_fit_multiblock_numpy(
                    X_blocks, y,
                    self.n_components_,
                    standardize=self.standardize
                )
                (self._B, self._W, self._P, self._Q, self._T,
                 self._block_weights, self._X_means, self._X_stds,
                 self._y_mean, self._y_std) = result
            else:
                result = _mbpls_fit_numpy(
                    X_blocks[0], y,
                    self.n_components_,
                    standardize=self.standardize
                )
                (self._B, self._W, self._P, self._Q, self._T,
                 self._X_mean, self._X_std,
                 self._y_mean, self._y_std) = result

            # Store coefficients
            self.coef_ = self._B

        return self

    def predict(self, X):
        """Predict using the MB-PLS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or list of arrays
            Samples to predict. Must match the format used in fit().

        Returns
        -------
        y_pred : ndarray of shape (n_samples,) or (n_samples, n_targets)
            Predicted values.
        """
        # Handle single array or list of blocks
        if isinstance(X, list):
            X_blocks = [np.asarray(x) for x in X]
        else:
            X = np.asarray(X)
            X_blocks = [X]

        if self.backend == 'jax':
            import jax.numpy as jnp

            _, mbpls_predict_jax = _get_cached_jax_mbpls()

            X_jax = jnp.asarray(X_blocks[0])
            y_pred = mbpls_predict_jax(
                X_jax, self._B,
                self._X_mean, self._X_std,
                self._y_mean, self._y_std
            )
            y_pred = np.asarray(y_pred)
        else:
            if self._is_multiblock:
                y_pred = _mbpls_predict_multiblock_numpy(
                    X_blocks, self._B,
                    self._X_means, self._X_stds,
                    self._y_mean, self._y_std
                )
            else:
                y_pred = _mbpls_predict_numpy(
                    X_blocks[0], self._B,
                    self._X_mean, self._X_std,
                    self._y_mean, self._y_std
                )

        # Flatten if single target
        if y_pred.ndim > 1 and y_pred.shape[1] == 1:
            y_pred = y_pred.ravel()

        return y_pred

    def transform(self, X):
        """Transform X to latent space.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or list of arrays
            Samples to transform.

        Returns
        -------
        T : ndarray of shape (n_samples, n_components)
            Latent variables (scores).
        """
        if isinstance(X, list):
            X_blocks = [np.asarray(x) for x in X]
        else:
            X = np.asarray(X)
            X_blocks = [X]

        if self.backend == 'jax':
            import jax.numpy as jnp

            X_jax = jnp.asarray(X_blocks[0])
            X_centered = (X_jax - self._X_mean) / self._X_std
            return np.asarray(X_centered @ self._W)
        else:
            if self._is_multiblock:
                return _mbpls_transform_multiblock_numpy(
                    X_blocks, self._W,
                    self._X_means, self._X_stds,
                    self._block_sizes
                )
            else:
                return _mbpls_transform_numpy(
                    X_blocks[0], self._W,
                    self._X_mean, self._X_std
                )

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        return {
            'n_components': self.n_components,
            'method': self.method,
            'standardize': self.standardize,
            'max_tol': self.max_tol,
            'backend': self.backend,
        }

    def set_params(self, **params):
        """Set the parameters of this estimator.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : MBPLS
            Estimator instance.
        """
        for key, value in params.items():
            setattr(self, key, value)
        return self

    def __repr__(self):
        """Return string representation."""
        params = [f"n_components={self.n_components}"]
        if self.method != 'NIPALS':
            params.append(f"method='{self.method}'")
        if not self.standardize:
            params.append("standardize=False")
        if self.backend != 'numpy':
            params.append(f"backend='{self.backend}'")
        return f"MBPLS({', '.join(params)})"
